import pyboolnet.prime_implicants
import pystablemotifs as sm
import copy
import itertools as it
import random
from collections import namedtuple

def fixed_implies_implicant(fixed,implicant):
    """Returns True if and only if the (possibly partial) state "fixed" implies
    the implicant.

    Parameters
    ----------
    fixed : partial state dictionary
        State (or partial state) representing fixed variable states.
    implicant : partial state dictionary
        State (or partial state) representing the target implicant.

    Returns
    -------
    bool
        True if and only if the implicant is in the logical domain of influence
        of the fixed (partial) state.

    """
    rval = True
    for k,v in implicant.items():
        if not k in fixed:
            rval = False
            break
        elif fixed[k] != v:
            rval = False
            break
    return rval

def fixed_excludes_implicant(fixed,implicant):
    """Returns True if and only if the (possibly partial) state "fixed"
    contradicts the implicant.

    Parameters
    ----------
    fixed : partial state dictionary
        State (or partial state) representing fixed variable states.
    implicant : partial state dictionary
        State (or partial state) representing the target implicant.

    Returns
    -------
    bool
        True if and only if the implicant contradicts the logical domain of
        influence of the fixed (partial) state.

    """
    rval = True
    for k,v in implicant.items():
        if not k in fixed:
            continue
        elif fixed[k] != v:
            rval = False
            break
    return not rval

def logical_domain_of_influence(partial_state,primes,implied_hint=None,contradicted_hint=None):
    """Computes the logical domain of influence (LDOI) (see Yang et al. 2018).
    In general, the LDOI is a subset of the full domain of influence (DOI), but it
    is much more easily (and quickly) computed.

    Parameters
    ----------
    partial_state : partial state dictionary
        pyboolnet implicant that defines fixed nodes.
    primes : pyboolnet primes dictionary
        Update rules.
    implied_hint : partial state dictionary
         Known subset of the LDOI; used during optimization.
    contradicted_hint : partial state dictionary
        Known subset of the contradiction boundary; used during optimization.

    Returns
    -------
    implied : partial state dictionary
        The logical domain of influence.
    contradicted : partial state dictionary
        The contradiction boundary.

    """

    if implied_hint is None:
        implied_hint = {}
    if contradicted_hint is None:
        contradicted_hint = {}

    fixed = partial_state.copy() # fixed will be partial_state + its LDOI
    fixed.update(implied_hint)
    implied = implied_hint # the LDOI
    contradicted = contradicted_hint # states implied by partial_state that contradict partial_state
    primes_to_search = primes.copy()

    for k in implied_hint: del primes_to_search[k]
    for k in contradicted_hint: del primes_to_search[k]

    while True:
        states_added = False
        deletion_list = []
        for k,v in primes_to_search.items():
            kfixed = False
            for i in [0,1]:
                if kfixed: break
                for p in v[i]:
                    if kfixed: break
                    if fixed_implies_implicant(fixed,p):
                        kfixed = True
                        deletion_list.append(k)
                        states_added = True
                        if k in fixed:
                            if fixed[k] == i: implied[k] = i
                            else: contradicted[k] = i
                        else:
                            implied[k] = i
                            fixed[k] = i
        for k in set(deletion_list): del primes_to_search[k]
        if not states_added or len(primes_to_search) == 0: break
    return implied, contradicted

def domain_of_influence(partial_state,primes,implied_hint=None,contradicted_hint=None,
    max_simulate_size=20,max_stable_motifs=10000,MPBN_update=False):
    """Computes the domain of influence (DOI) of the seed set. (see Yang et al. 2018)

    Parameters
    ----------
    partial_state : partial state dictionary
        pyboolnet implicant that defines fixed nodes (seed set).
    primes : pyboolnet primes dictionary
        Update rules.
    implied_hint : partial state dictionary
        Known subset of the DOI; used during optimization.
    contradicted_hint : partial state dictionary
        Known subset of the contradiction boundary; used during optimization.
    max_simulate_size : int
        Maximum number of variables for which to brute-force build a state
        transition graph (the default is 20).
    max_stable_motifs : int
        Maximum number of output lines for pyboolnet to process from the
        AspSolver (the default is 10000).
    MPBN_update : bool
        Whether MBPN update is used instead of general asynchronous update
        (see Pauleve et al. 2020)(the default is False).
    Returns
    -------
    data : namedtuple
        A namedtuple that contains the entries listed below.
    implied : partial state dictionary
        Nodes that are certain to be in the domain of influence.
    contradicted : partial state dictionary
        The contradiction boundary.
    possibly_implied : partial state dictionary
        Nodes that are possibly in the domain of influence.
    possibly_contradicted : partial state dictionary
        Nodes that are possibly in the contradiction boundary.
    attractor_repertoire : AttractorRepertoire
        The class that stores information about attractors.
    """
    # optional optimization values
    if implied_hint is None:
        implied_hint = {}
    if contradicted_hint is None:
        contradicted_hint = {}

    implied = implied_hint # the DOI
    contradicted = contradicted_hint # states implied by partial_state that contradict partial_state
    unknown = {}
    unknown_contra = {}
    primes_to_search = copy.deepcopy(primes)
    for node in implied_hint: del primes_to_search[node]
    for node in contradicted_hint: del primes_to_search[node]

    # fixed will be partial_state + its LDOI
    fixed = partial_state.copy()
    LDOI,LDOI_contra = logical_domain_of_influence(partial_state,primes)
    fixed.update(LDOI)
    implied.update(LDOI)
    contradicted.update(LDOI_contra)

    # reducing the primes by the LDOI
    primes_to_search, ps = sm.reduction.reduce_primes(fixed,primes_to_search)

    # Adding the seed nodes that are not in the LDOI as sinks in the reduced network
    sink_nodes = {}
    for node,value in fixed.items():
        if node not in LDOI:
            sink_nodes[node] = value

    # Adding back the simplifed rules for the sink_nodes
    rules_to_add = {}
    for node in sink_nodes:
        rules_to_add[node] = copy.deepcopy(primes[node])
        pyboolnet.prime_implicants._substitute(rules_to_add,node,fixed)
    primes_to_search.update(rules_to_add)
    primes_to_search = sm.reduction.simplify_primes(primes_to_search)

    # Finding the attractors of this modified network
    if MPBN_update == True:
        attractor_dict_list = pyboolnet.trap_spaces.compute_trap_spaces(primes_to_search, "min", max_output=10000)
        sorted_attractor_dict_list = []
        for attractor in attractor_dict_list:
            for node in primes_to_search:
                if node not in attractor:
                    attractor[node] = 'X'
            sorted_attractor_dict_list.append(dict(sorted(attractor.items())))
        ar = sorted_attractor_dict_list
    else:
        ar = sm.AttractorRepertoire.from_primes(primes_to_search,max_simulate_size=max_simulate_size,
            max_stable_motifs=max_stable_motifs,MPBN_update=MPBN_update)
        attractor_dict_list = []
        for attractor in ar.attractors:
            attractor_dict_list.append(attractor.attractor_dict)

    # Determining which node values are shared by all attractors in the reduction
    if len(attractor_dict_list) == 0:   # there is no attractor
        print("Unable to properly count attractors.")
        return

    att = attractor_dict_list[0]  # there is at least one attractor.
    for node in att:
        value = None
        value_implied = None
        value_unknown = None
        if node in LDOI:    # if the node is in the LDOI, it is in the DOI.
            implied[node] = att[node]
            continue    # this node is in the DOI. Move on to the next.
        elif att[node] == 'X':
            continue    # this node is not in the DOI. Move on to the next.
        elif att[node] == '?' or att[node] == '!':
            value_unknown = True
        else:   # in case the node has a Boolean value
            value = att[node]

        for other in attractor_dict_list: # check the other attractors
            if node in other:
                if other[node] == 'X':
                    value_implied = False
                    break
                elif other[node] == '?' or other[node] == '!':
                    value_unknown = True
                else:   # the node in the other attractor has a Boolean value.
                    if value == None:
                        value = other[node]
                    elif int(value) == int(other[node]):
                        continue
                    else:
                        value_implied = False
                        break
            else:   # the node is not in the other attractor.
                value_implied = False
                break

        if value_implied == False:
            continue    # this node is not in the DOI, move on to the next node.
        elif value_unknown == True:
            if value == None:
                unknown[node] = 'Unknown'
            else:
                unknown[node] = int(value)
        else:
            implied[node] = int(value)

    for node in sink_nodes:
        if node in implied:
            if sink_nodes[node] != implied[node]:
                contradicted[node] = implied[node]
                del implied[node]
        elif node in unknown:
            if sink_nodes[node] != unknown[node]:
                unknown_contra[node] = unknown[node]

    doi_data = namedtuple('doi_data',
        'implied contradicted possibly_implied possibly_contradicted attractor_repertoire')
    data = doi_data(dict(sorted(implied.items())), dict(sorted(contradicted.items())),
        dict(sorted(unknown.items())), dict(sorted(unknown_contra.items())), ar)

    return data

def single_drivers(target,primes):
    """Finds all 1-node (logical) drivers of target under the rules given
    by primes.

    Parameters
    ----------
    target : partial state dictionary
        pyboolnet implicant that defines target fixed node states.
    primes : pyboolnet primes dictionary
        Update rules.

    Returns
    -------
    list of length-1 dictionaries
        Each dictionary describes a single node state that contains the target
        in its logical domain of influence.

    """
    drivers = []
    for k in primes:
        for val in [0,1]:
            ds = {k:val}
            ldoi,contra = logical_domain_of_influence(ds,primes)
            if all([kk in ldoi for kk in target]):
                drivers.append(ds)
    return drivers

def all_drivers_of_size(driver_set_size,target, primes, external_search_vars=None,internal_search_vars=None):
    """Short summary.

    Parameters
    ----------
    driver_set_size : int
        The number of driver nodes to try to find.
    target : partial state dictionary
        pyboolnet implicant that defines target fixed node states.
    primes : pyboolnet primes dictionary
        Update rules.
    external_search_vars : set of str variable names
        Node set not in target to consider as potential drivers. If None,
        then all nodes not fixed in target (the default is None).
    internal_search_vars : set of str variable names
        Node set in target to consider as potential drivers. If None, all
        nodes in partial state (the default is None).

    Returns
    -------
    driver_sets : list of partial state dictionaries
        Each state dictionary in the list drives target.

    """

    if internal_search_vars is None:
        internal_search_vars = set(target.keys())
    if external_search_vars is None:
        external_search_vars = set(x for x in primes if not x in target)

    driver_sets = []

    search_vars = internal_search_vars | external_search_vars
    for driver_vars in it.combinations(search_vars,driver_set_size):
        # Any internal nodes must be in their target state to be drivers
        internal_driver_dict = {k:target[k] for k in driver_vars if k in target}
        out_keys = [k for k in driver_vars if not k in target]
        for s in it.product([0,1],repeat=len(out_keys)):
            external_driver_dict = {k:s for k,s in zip(out_keys,s)}
            driver_dict = {**internal_driver_dict, **external_driver_dict}
            is_candidate=True
            # The driver set we're looking at has too many nodes if we already
            # found a driver set that is a subset of it.
            for known_driver_set in driver_sets:
                if driver_dict.items() >= known_driver_set.items():
                    is_candidate = False
                    break
            if not is_candidate:
                continue

            implied,contradicted = logical_domain_of_influence(driver_dict,primes)
            fixed = {**implied, **driver_dict}

            target_stabilized = True
            for k,v in target.items():
                if not k in fixed or not fixed[k] == v:
                    target_stabilized = False
                    break

            if target_stabilized:
                driver_sets.append(driver_dict)
    return driver_sets

def internal_drivers(target,primes,max_drivers=None):
    """Find internal (logical) driver nodes of target through brute-force.

    Parameters
    ----------
    target : partial state dictionary
        pyboolnet implicant that defines target fixed node states.
    primes : pyboolnet primes dictionary
        Update rules.
    max_drivers : int
        Maximum size of driver set to consider. If None, is set to the size of
        the parital state (as this is always sufficient to achieve the target)
        (the default is None).

    Returns
    -------
    driver_sets : list of partial state dictionaries
        Each state dictionary in the list drives target. These are sorted
        by length (smallest first).

    """
    if max_drivers is None:
        max_drivers = len(target) # The target itself is always its own driver

    driver_sets = []

    for driver_set_size in range(1,max_drivers+1):
        driver_sets += all_drivers_of_size(driver_set_size,target,primes,external_search_vars=set())

    if len(driver_sets) == 0:
        driver_sets.append(target)

    return sorted(driver_sets, key = lambda x: len(x))


def minimal_drivers(target,primes,max_drivers=None):
    """Finds smallest set(s) of (logical) driver nodes of target through
    brute-force. Unlike minimal_drivers, we are not limited to internal drivers
    nodes.

    Parameters
    ----------
    target : partial state dictionary
        pyboolnet implicant that defines target fixed node states.
    primes : pyboolnet primes dictionary
        Update rules.
    max_drivers : int
        Maximum size of driver set to consider. If None, is set to the size of
        the parital state (as this is always sufficient to achieve the target)
        (the default is None).

    Returns
    -------
    driver_sets : list of partial state dictionaries
        Each state dictionary in the list drives target. These are sorted
        by length (smallest first).

    """

    if max_drivers is None:
        max_drivers = len(target) # The target itself is always its own driver

    driver_sets = []
    out_vars = set(x for x in primes if not x in target)
    for driver_set_size in range(1,max_drivers+1):
        if len(driver_sets) > 0:
            break
        driver_sets += all_drivers_of_size(driver_set_size,target,primes,external_search_vars=out_vars)

    if len(driver_sets) == 0:
        driver_sets.append(target)

    return sorted(driver_sets, key = lambda x: len(x))

def knock_to_partial_state(target,primes,min_drivers=1,max_drivers=None,forbidden=None):
    """Find all partial states in primes that drive the target. Do not consider
    nodes in the forbidden list.

    Parameters
    ----------
    target : partial state dictionary
        pyboolnet implicant that defines target fixed node states.
    primes : pyboolnet primes dictionary
        Update rules.
    min_drivers : int
        Minimum size of driver set to consider. (the default is 1).
    max_drivers : int
        Maximum size of driver set to consider. If None, is set to the size of
        the parital state (as this is always sufficient to achieve the target)
        (the default is None).
    forbidden : set of str variable names
        Variables to be considered uncontrollable (the default is None).

    Returns
    -------
    knocked_nodes : list of partial state dictionaries
        Each state dictionary in the list drives target. Supersets of
        previously considered dictionaries are excluded.

    """
    if max_drivers is None:
        max_drivers = len(primes) - len(target)
    if forbidden is None:
        forbidden = set()

    knocked_nodes = []

    internal_search = set([k for k in target if not k in forbidden])
    external_search = set([k for k in primes if not k in forbidden and not k in target])

    for n in range(min_drivers,max_drivers+1):
        found = all_drivers_of_size(n,target,primes,
            internal_search_vars=internal_search,external_search_vars=external_search)
        for x in found:
            add_x = True
            for y in knocked_nodes:
                if x.items() >= y.items():
                    add_x = False
                    break
            if add_x:
                knocked_nodes.append(x)

        n += 1
    return knocked_nodes

################################################################################
# Below are Greedy Random Adaptive Search Program (GRASP) methods.
# These were developed in Yang et al. 2018.
################################################################################

def _initial_GRASP_candidates(target,primes,forbidden):
    """Helper function for GRASP driver search. Constructs initial candidates
    for driver nodes.

    Parameters
    ----------
    target : partial state dictionary
        pyboolnet implicant that defines target fixed node states.
    primes : pyboolnet primes dictionary
        Update rules.
    forbidden : set of str variable names
        Variables to be considered uncontrollable (the default is None).

    Returns
    -------
    candidates : list of partial state dictionaries
        List of variable states that can potentially lead to the target.

    """
    if forbidden is None:
        candidate_vars = list(primes.keys())
    else:
        candidate_vars = [k for k in primes if not k in forbidden]
    candidates = []
    for st in [0,1]:
        candidates += [{k:st} for k in candidate_vars]
    return candidates

def _GRASP_default_scores(target,primes,candidates):
    """Helper function for GRASP driver search. Scores candidate driver nodes.

    Parameters
    ----------
    target : partial state dictionary
        pyboolnet implicant that defines target fixed node states.
    primes : pyboolnet primes dictionary
        Update rules.
    candidates : list of partial state dictionaries
        List of variable states that can potentially lead to the target.

    Returns
    -------
    scores : list of ints
        Logical domain of influence sizes for individual node states. If the
        node leads to a contradiction, the score will become the negative of the
        largest LDOI size. Scores are ordered in the same order as the
        candidates list.

    """
    scores = []
    m = 0 # will be the size of largest LDOI
    for candidate in candidates:
        imp,con = logical_domain_of_influence(candidate,primes)

        sc = (1+len([x for x in imp if x in target and imp[x]==target[x]])) * len(imp)
        if sc > m:
            m = sc

        if any(k in target and target[k] == con[k] for k in con):
            scores.append(-1*sc - 1)
        else:
            scores.append(sc)

    # score will be len(imp) if no contradictions in LDOI
    # but will be len(imp) - m otherwise
    scores = [x - int(x < 0)*(m+2*x+1) for x in scores]

    return scores

def _construct_GRASP_solution(target,primes,candidates,scores):
    """Helper funciton for GRASP driver search. Constructs individual driver set
    using the GRASP search method.

    Parameters
    ----------
    target : partial state dictionary
        pyboolnet implicant that defines target fixed node states.
    primes : pyboolnet primes dictionary
        Update rules.
    candidates : list of partial state dictionaries
        List of variable states that can potentially lead to the target.
    scores : list of ints
        Logical domain of influence sizes for individual node states. If the
        node leads to a contradiction, the score will become the negative of the
        largest LDOI size. Scores are ordered in the same order as the
        candidates list.

    Returns
    -------
    partial state dictionary
        A partial state that contains the target in its LDOI. If no such partial
        state is found, returns an empty dictionary instead.

    """
    solution = {}
    alpha = random.random()

    #pass_score = alpha * (max(scores) - min(scores)) + min(scores)
    scores_pos = [x for x in scores if x > 0] + [0] # add 0 to avoid min/max errors
    pass_score = alpha * (max(scores_pos) - min(scores_pos)) + min(scores_pos)

    RCL = [x for i,x in enumerate(candidates) if scores[i] >= pass_score]

    imp_old = None
    con_old = None

    while len(RCL) > 0:
        s = random.choice(RCL)
        new_solution = solution.copy()
        new_solution.update(s)

        imp,con = logical_domain_of_influence(new_solution,primes,implied_hint=imp_old,contradicted_hint=con_old)

        if any(k in target and target[k] != con[k] for k in con):
            RCL = [x for x in RCL if x != s]
            continue

        if target.items() <= imp.items():
            return new_solution

        imp_old = imp
        con_old = con
        solution = new_solution
        RCL = [x for x in RCL if not next(iter(x.keys())) in solution]

    return {}

def _local_GRASP_reduction(solution,target,primes):
    """A helper funciton for GRASP driver search. Reduces valid solutions to
    attempt to remove redundancies.

    Parameters
    ----------
    solution : partial state dictionary
        Solution to be reduced; must contain the target in its LDOI.
    target : partial state dictionary
        pyboolnet implicant that defines target fixed node states.
    primes : pyboolnet primes dictionary
        Update rules.

    Returns
    -------
    partial state dictionary
        Reduced solution that also contains target in its LDOI.

    """
    keylist = list(solution.keys())
    random.shuffle(keylist)
    old_solution = {x:solution[x] for x in keylist}

    for k in keylist:
        new_solution = {x:v for x,v in old_solution.items() if x != k}
        if len(new_solution) == 0:
            return old_solution

        imp,con = logical_domain_of_influence(new_solution, primes)
        if target.items() <= imp.items():
            old_solution = new_solution.copy()

    return old_solution

def GRASP(target, primes, GRASP_iterations, forbidden=None, GRASP_scores=_GRASP_default_scores):
    """Search for drivers of target in primes using the method of Yang et al. 2018.

    Parameters
    ----------
    target : partial state dictionary
        pyboolnet implicant that defines target fixed node states.
    primes : pyboolnet primes dictionary
        Update rules.
    GRASP_iterations : int
        The number of times to run the GRASP method.
    forbidden : set of str variable names
        Variables to be considered uncontrollable (the default is None).
    GRASP_scores : function
        Function to score candiates (the default is _GRASP_default_scores; see
        that function for required inputs and outputs of the scoring function).

    Returns
    -------
    solutions : list of partial state dictionaries
        Each partial set dictionary represents a driver set whose LDOI contains
        the target.

    """
    solutions = []
    candidates = _initial_GRASP_candidates(target,primes,forbidden)
    scores = GRASP_scores(target,primes,candidates)
    for iter in range(GRASP_iterations):
        solution_big = _construct_GRASP_solution(target,primes,candidates,scores)
        solution = _local_GRASP_reduction(solution_big,target,primes)
        if not solution is None and len(solution) > 0 and not solution in solutions:
            solutions.append(solution)

    return solutions



import pystablemotifs.succession as sm_succession
import pystablemotifs.Attractor as Attractor
import itertools as it
import networkx as nx
import pyboolnet.prime_implicants

class AttractorRepertoire:
    """The class that stores information about attractors. Initialize using
    either from_primes or from_succession_diagram.

    Attributes
    ----------
    succession_diagram : succession.SuccessionDiagram
        Succession diagram summarizing the stable motif structure of the model.
    attractors : list of Attractor.Attractor
        List of (possible) attractors in the model.
    reduction_attractors : dictionary
        A dictionary with integer keys that correspond to the
        succession_diagram.digraph nodes. The dictionary values are lists of
        Attractor.Attractor objects that correspond to attractors that exist in
        the region of statespace corresponding to the reduced network
        represented by the key in the succession diagram.
    fewest_attractors : int
        A lower bound on the number of attractors in the model.
    most_attractors : int
        An upper bound on the number of attractors in the model.
    primes : pyboolnet primes dictionary
        The model rules.
    succession_digraph : networkx digraph
        Networkx digraph representation of the succession_diagram object. If
        AttractorRepertoire.simplify_diagram, it is equivalent to
        AttractorRepertoire.succession_diagram.digraph. Otherwise, several of its
        nodes may be contracted (depending on input parameters).
    attractor_equivalence_classes : list
        List of attractor equivalence classes. Each item is a dictionary with keys
        'states', 'attractors', and 'reductions'. The 'states' value is a dictionary
        of variable values that all attractors in the class share. The 'attractors'
        value is a list of Attractor objects (i.e., a sublist of self.attractors);
        all attractors in this list have all relevant nodes equivalently characterized.
        The 'reductions' value is a list of reduction_attractor keys that collectively
        contain all the attractors in the class (and therefore cannot differ in any
        relevant node).
    relevant_nodes : list
        List of nodes that are "relevant", i.e., if trap spaces differ in the values
        of these variables, then the corresponding succession diagram nodes and
        attractors will not be merged.

    """

    def __init__(self):
        self.succession_diagram = None
        self.attractors = []
        self.reduction_attractors = {}
        self.fewest_attractors = None
        self.most_attractors = None
        self.primes = None

        self.succession_digraph = None
        self.attractor_equivalence_classes = None
        self.relevant_nodes = None

    @classmethod
    def from_primes(cls,primes,max_simulate_size=20,max_stable_motifs=10000,MPBN_update=False):
        """Build the succession diagram and attractor repertoire from pyboolnet
        formatted update rules rules.

        Parameters
        ----------
        primes : pyboolnet primes dictionary
            The model rules.
        max_simulate_size : int
            Maximum number of variables for which to brute-force build a state
            transition graph (the default is 20).
        max_stable_motifs : int
            Maximum number of output lines for pyboolnet to process from the
            AspSolver (the default is 10000).
        MPBN_update : bool
            Whether MBPN update is used instead of general asynchronous update
            (see L Pauleve, J Kolcak, T Chatain, S Haar, "Reconciling qualitative, abstract, and scalable modeling of biological networks." Nat. Com. vol. 11, no. 4256 (2020))
            (the default is False).

        Returns
        -------
        AttractorRepertoire
            AttractorRepertoire object for the input primes.

        """
        x = cls()
        x.primes = primes
        x.analyze_system(primes,max_simulate_size=max_simulate_size,max_stable_motifs=max_stable_motifs,MPBN_update=MPBN_update)
        x.simplify_diagram([], merge_equivalent_reductions = False)
        return x

    @classmethod
    def from_succession_diagram(cls,succession_diagram):
        """Build the succession diagram and attractor repertoire from a
        precomputed succession diagram.

        Parameters
        ----------
        succession_diagram : succession.SuccessionDiagram
            Succession diagram summarizing the stable motif structure of the model.

        Returns
        -------
        AttractorRepertoire
            AttractorRepertoire object for the input succession diagram.

        """
        x = cls()
        x.succession_diagram = succession_diagram
        x.primes = succession_diagram.unreduced_primes
        x._get_attractors_from_succession_diagram()
        x._count_attractors()
        x.simplify_diagram([], merge_equivalent_reductions = False)
        return x


    def _get_attractors_from_succession_diagram(self):
        """Extract attractors from the succession diagram of the model.

        """
        for ri, reduction in self.succession_diagram.motif_reduction_dict.items():
            if reduction.terminal == "no": continue

            self.reduction_attractors[ri] = []

            duplicate = False # have we added the attractors for this reduciton yet?
            for attractor in self.attractors:
                if attractor.logically_fixed_nodes == reduction.logically_fixed_nodes:
                    duplicate = True
                    attractor.add_reduction(reduction)
                    self.reduction_attractors[ri].append(attractor)

            if duplicate: continue

            for id,att in enumerate(reduction.attractor_dict_list):
                new_attractor = Attractor(reduction,id)
                self.attractors.append(new_attractor)
                self.reduction_attractors[ri].append(new_attractor)

    def _count_attractors(self):
        """Place upper and lower bounds on the number of attractors.

        """
        self.fewest_attractors = 0
        self.most_attractors = 0
        for attractor in self.attractors:
            if attractor.guaranteed:
                self.fewest_attractors += 1
            # else: self.fewest_attractors += 0

            if attractor.explored:
                self.most_attractors += 1
            else:
                if attractor.representative[0].deletion_no_motif_attractors is not None:
                    self.most_attractors += len(attractor.representative[0].deletion_no_motif_attractors)
                else:
                    # ludicrously conservative upper bound; assumes STG is all 2-cycles
                    self.most_attractors += 2**(attractor.n_unfixed - 1)

    def analyze_system(self,primes,max_simulate_size=20,max_stable_motifs=10000,MPBN_update=False):
        """Build and process the succession diagram for the model.

        Parameters
        ----------
        primes : pyboolnet primes dictionary
            The model rules.
        max_simulate_size : int
            Maximum number of variables for which to brute-force build a state
            transition graph (the default is 20).
        max_stable_motifs : int
            Maximum number of output lines for pyboolnet to process from the
            AspSolver (the default is 10000).
        MPBN_update : bool
            Whether MBPN update is used instead of general asynchronous update
            (see Pauleve et al. 2020)(the default is False).

        """
        self.succession_diagram = sm_succession.build_succession_diagram(primes,max_simulate_size=max_simulate_size,max_stable_motifs=max_stable_motifs,MPBN_update=MPBN_update)
        self._get_attractors_from_succession_diagram()
        self._count_attractors()

    def simplify_diagram(self, projection_nodes, merge_equivalent_reductions = True, keep_only_projection_nodes = False, condense_simple_paths = False):
        """Simplify the succession diagram for the model. This is done in two ways.
        First, variables can be designated ignorable using the projection_nodes
        parameter. If keep_only_projection_nodes is False, these variables are
        ignorable, otherwise, all other nodes are ignorable. When
        merge_equivalent_reductions is True, all nodes of the succession diagram
        that correspond to trap spaces whose fixed variables differ only in ignorable
        variables are contracted (in the graph theory sense). After this process,
        if condense_simple_paths is True, then all succession diagram nodes with
        in-degree equal to one are contracted with their parent node. This function
        constructs the succession_digraph and attractor_equivalence_classes attributes,
        which are described in the class documentation.

        Parameters
        ----------
        projection_nodes : list of variable names
            These nodes will be ignored if keep_only_projection_nodes is False
            (default); otherwise, all nodes except these will be ignored.
        merge_equivalent_reductions : bool
            Whether to contract succession diagram nodes whose reductions differ
            only in ignorable nodes.
        keep_only_projection_nodes : bool
            Whether projection_nodes specifies non-ignorable nodes.
        condense_simple_paths : bool
            Whether to contract nodes with in-degree one.
        """

        if not keep_only_projection_nodes:
            keep = set(self.primes.keys()) - set(projection_nodes)
            ignore = set(projection_nodes)
        else:
            keep = set(projection_nodes)
            ignore = set(self.primes.keys()) - set(projection_nodes)

        G = self.succession_diagram.digraph.copy()

        # Merge equivalent nodes
        if merge_equivalent_reductions:
            for u,v in it.combinations(self.succession_diagram.digraph.nodes(),2):
                if u not in G.nodes(): continue # we've already merged u

                ru = self.succession_diagram.motif_reduction_dict[u]
                rv = self.succession_diagram.motif_reduction_dict[v]
                if pyboolnet.prime_implicants.primes_are_equal(ru.reduced_primes,rv.reduced_primes):
                    rud = ru.logically_fixed_nodes
                    rvd = rv.logically_fixed_nodes
                    if all([rud[k]==rvd[k] for k in keep if k in rud and k in rvd]):
                        G = nx.contracted_nodes(G,u,v,self_loops=False)

        # If a node v has out degree = 1, then merge it into its only child u
        if condense_simple_paths:
            old_N = float('inf')
            while G.number_of_nodes() < old_N:
                old_N = G.number_of_nodes()
                for v in G.nodes():
                    if G.out_degree(v) == 1:
                        u = next(G.successors(v))
                        G = nx.contracted_nodes(G,u,v,self_loops=False)


        self.succession_digraph = G

        self.attractor_equivalence_classes = []
        for a in self.attractors:
            keys = set()
            for n in self.succession_digraph.nodes():
                if n not in self.reduction_attractors: continue
                for b in self.reduction_attractors[n]:
                    if all([b.attractor_dict[k] == a.attractor_dict[k] for k in keep]):
                        keys.add(n)
                        break

            merged = False
            for c in self.attractor_equivalence_classes:
                if c['states'].items() <= a.attractor_dict.items():
                    c['attractors'].append(a)
                    c['reductions'] |= keys
                    merged = True
            if not merged:
                self.attractor_equivalence_classes.append({'states':{k:v for k,v in a.attractor_dict.items() if k in keep},'attractors':[a], 'reductions':keys})

        self.relevant_nodes = keep

    def summary(self):
        """Prints a summary of the attractors to standard output.

        """
        if self.fewest_attractors == 0:
            print("Unable to properly count attractors.")
        elif self.fewest_attractors == self.most_attractors:
            if self.fewest_attractors == 1:
                print("There is 1 attractor.")
            else:
                print("There are",self.fewest_attractors,"attractors.")
        else:
            print("There are between",self.fewest_attractors,"and",self.most_attractors,"attractors.")
        for att in self.attractors:
            print(att.attractor_dict)
            print()

    def reprogram_to_trap_spaces(self,logically_fixed,
        target_method='history',driver_method='internal',
        max_drivers=None,GRASP_iterations=None,GRASP_score_override=None):
        """Find driver sets that lead to fixing the node states specified.

        Parameters
        ----------
        logically_fixed : partial state dictionary
            Targeted fixed nodes.
        target_method : str
            Either 'history' or 'merge'; see Notes below for details.
        driver_method : str
            Either 'internal', 'minimal', or 'GRASP' see Notes below for details.
        max_drivers : int
            Maximum number of driver nodes to consider (not used in GRASP methods).
            If none, the upper limit is given by the number of free variables
            (the default is None).
        GRASP_iterations : int
            Number of times to construct GRASP driver sets; only used in GRASP
            methods. If none, the number of iterations is chosen based on the
            network size (the default is None).
        GRASP_score_override : function
            Optional heuristic score function override (see drivers.GRASP
            for details). Only used in GRASP methods (the default is None).

        Returns
        -------
        list
            Control strategies found; interpretation depends on method selected
            See Notes below for details.

        Notes
        -----
        The various combinations of target_method and driver_method options result
        in different control strategies, which are outlined below.

        target_method = history, driver_method = internal:
        Finds all shortest stable motif histories that result in the target node states
        being logically fixed. Each stable motif is searched for internal driver nodes.
        The resulting internal drivers are combined into a single  control set. The
        return value consists of all such control sets for all  stable motif histories.
        Each control set eventually becomes self-sustaining.

        target_method = history, driver_method = minimal:
        Similar to the history method, except the search for stable motif drivers
        includes external driver nodes for the motif and does not extend to driver sets
        of larger size once one driver set has been found for a motif. Because the
        search includes external driver nodes, special care must be taken in interpreting
        the effect of the drivers, as their influence may impact the effect of motifs
        stabilizing. Thus, the control is only guaranteed to work if the interventions
        are temporary and implemented in the order specified by the motif history.

        For this reason, the output consists of lists of ordered interventions.
        Each element of the return value is a list of lists of dictionaries. Each
        element of the return value represents a control strategy. To implement such
        a strategy, select a dictionary from the first element of the strategy and
        fix the node states it specifies until their influence has propagated through
        the system. Then repeat this process iteratively for each element of the strategy
        list, in order. For example, if
        nonredundant_drivers = [ [[{'xD':1,'xE=1'}]], [[{'xA':1},{'xB':1}],[{'xC':1}]] ]
        then there are two control strategies available:
        1) fix xD=xE=1 temporarily and
        2) first fix either xA=1 or xB=1 temporarily, then fix xC=1 temporarily.

        target_method = history, driver_method = GRASP:
        The same as history, minimal, except external driver nodes are searched for
        using the GRASP algorithm using GRASP_iterations iterations.

        target_method = merge, driver_method = internal:
        Finds all shortest stable motif histories that result in the target node states
        being logically fixed. All node states in the motifs in the history are merged
        into a stable module dictionary. This is then searched for internal driver
        nodes. Each element of the return value is a dictionary corresponding to a
        control set. Each control set eventually becomes self-sustaining.

        target_method = merge, driver_method = minimal:
        Similar to the merge method, except the search for drivers is conducted over
        all nodes, not just those internal to the merged stable module. Furthermore,
        the search is truncated when a control set is found such that the search does
        not proceed to driver sets larger than the smallest found. Each element of
        the return value is a dictionary corresponding to a control set. The control
        sets are only guaranteed to result in activation of the target if they are
        temporary interventions.

        target_method = merge, driver_method = GRASP:
        The same as merge, minimal, except external driver nodes are searched for
        using the GRASP algorithm using GRASP_iterations iterations.

        """
        return self.succession_diagram.reprogram_to_trap_spaces(logically_fixed,
            target_method=target_method,driver_method=driver_method,
            max_drivers=max_drivers,
            GRASP_iterations=GRASP_iterations,GRASP_score_override=GRASP_score_override)



import pandas as pd
import random as rd
import numpy as np
from datetime import datetime
from uuid import uuid4
import os

class RandomBooleanNetworks:
    """Generator of random Boolean networks (RBN) and ensembles of RBN.
    The RandomBooleanNetworks class object is a Boolean model and stores
    information of how the Boolean model was generated. It has functions that
    generate ensembles of RBN by generating multiple RandomBooleanNetworks objects.

    Attributes
    ----------
    node_names : list of str
        List of variable names.
    node_inputs_dictionary : dictionary
        Each value is a (fixed order) list of the names of the nodes whose values
        are inputs into the key variable's update function.
    node_rules_binary_dictionary : dictionary
        Each value is a list of outputs for the key variable's update function,
        stored as a list in ascending order of the numerical representation of
        the input row.
    node_rules_decimal_dictionary : dictionary
        Decimal conversion of node_rules_binary_dictionary.
    node_rules_string_dictionary : dictionary
        BooleanNet (str) conversion of node_rules_binary_dictionary values.
    node_rules_string : str
        BooleanNet representation of update rules.
    random_boolean_type : str
        Descrpition of generative process. Currently only "Kauffman NK" is
        implemented.
    N : int
        Number of nodes in the Boolean network.
    random_boolean_Network_parameters : list
        For Kauffman NK generation -
            [K,p], where K is the in-degree and p is the bias. K is a positive
            integer less than or equal to N, and p is a float between 0 and 1
            (inclusive).
    random_seed : int
        Seed for random functions.
    filename : str
        Path to file where network data are stored. If None, no files are written.

    """

    def __init__(self):
        self.node_names = []
        self.node_inputs_dictionary = {}
        self.node_rules_binary_dictionary = {}
        self.node_rules_decimal_dictionary = {}
        self.node_rules_string_dictionary = {}
        self.node_rules_string = ""
        self.random_boolean_type = ""
        self.N=0
        self.random_boolean_Network_parameters= []
        self.random_seed = None
        self.filename = None

    def random_boolean_network(self,random_boolean_type,N,rbn_parameters,seed=None,filename=None):
        """Construct network using specified generative process.

        Parameters
        ----------
        random_boolean_type : str
            Descrpition of generative process. Currently only "Kauffman NK" is
            implemented.
        N : int
            Number of nodes in the Boolean network.
        random_boolean_Network_parameters : list
            For Kauffman NK generation -
                [K,p], where K is the in-degree and p is the bias. K is a positive
                integer less than or equal to N, and p is a float between 0 and 1
                (inclusive).
        random_seed : int
            Seed for random functions.
        filename : str
            Path to file where network data are stored. If None, no files are written.

        """
        self.random_boolean_type = random_boolean_type
        self.random_boolean_Network_parameters= rbn_parameters
        self.N=N
        if seed is not None:
            rd.seed(seed)
        self.node_names=["n"+str(i) for i in range(N)]

        if(random_boolean_type=="Kauffman NK"):
            K,p=self.random_boolean_Network_parameters
            self.node_inputs_dictionary={node_name:rd.sample(self.node_names,K) for node_name in self.node_names}
            self.node_rules_binary_dictionary={node_name:
                                         (rd.choices([1,0],weights=[p,1-p],k=np.power(2,len(node_input_list))),
                                          node_input_list)
                                         for node_name,node_input_list in self.node_inputs_dictionary.items()}
            if filename is not None:
                self.filename = filename
                self.node_rules_decimal_dictionary={node_name:
                                      (int("".join([str(i) for i in node_rule_input_list[0]]),2),
                                       node_rule_input_list[1])
                                      for node_name,node_rule_input_list in self.node_rules_binary_dictionary.items()}
                write_boolean_network_decimal(self.node_rules_decimal_dictionary,self.filename)

    def random_boolean_network_Rules(self):
        """Generate various conversions of the node_rules_binary_dictionary
        attribute.

        """
        self.node_rules_string_dictionary=String_Rules_From_Binary(self.node_rules_binary_dictionary)
        self.node_rules_string=""
        for n,f in self.node_rules_string_dictionary.items():
             self.node_rules_string=self.node_rules_string+n+" *= "+f+"\n"
        return(self.node_rules_string)

def write_boolean_network_decimal(node_rules_decimal_dictionary,filename):
    """Write the decimal format of the Boolean rules to file.

    Parameters
    ----------
    node_rules_decimal_dictionary : dictionary
        Update rule truth table in decimal format.
    filename : str
        Path to file for csv output of the truth table.

    """
    df_boolean_model=pd.DataFrame.from_dict(node_rules_decimal_dictionary,orient='index')
    df_boolean_model.to_csv(filename, header=False)

def read_boolean_network_decimal(filename):
    """Imports rules from csv in decimal format.

    Parameters
    ----------
    filename : str
        Path to csv from which to import decimal-formatted rules.

    Returns
    -------
    str
        Rules in BooleanNet format.

    """
    df_dict=pd.read_csv(filename,header=None,index_col=0).to_dict('index')
    node_rules_decimal_dictionary={key:(element[1],[x[1:-1] for x in element[2].strip('][').split(', ')])  for key,element in df_dict.items()}
    node_rules_binary_dictionary=Binary_Rules_From_Decimal(node_rules_decimal_dictionary)
    node_rules_string_dictionary=String_Rules_From_Binary(node_rules_binary_dictionary)

    rules=""
    for n,f in node_rules_string_dictionary.items():
        rules=rules+n+" *= "+f+"\n"
    return(rules)

def Binary_Rules_From_Decimal(node_rules_decimal_dictionary):
    """Construct Binary format rules from decimal format rules.

    Parameters
    ----------
    node_rules_decimal_dictionary : dictionary
        Rules in decimal format to convert.

    Returns
    -------
    dictionary
        Binary rules dictionary.

    """
    node_rules_binary_dictionary={node_name:
                                 (Binary_Rule_From_Decimal(node_rule_input_list[0],node_rule_input_list[1]),
                                            node_rule_input_list[1])
                                 for node_name,node_rule_input_list in node_rules_decimal_dictionary.items()}
    return(node_rules_binary_dictionary)

def Binary_Rule_From_Decimal(node_rule_decimal,node_input_list):
    """Convert single decimal rule to its binary form.

    Parameters
    ----------
    node_rule_decimal : int
        Decimal form of a truth table's output column.
    node_input_list : list of str
        Variable names that correspond to each column of the truth table.

    Returns
    -------
    list of int
        Binary rule list corresponding to an output column of a truth table.

    """
    binary_rule=[int(binary) for binary in np.binary_repr(node_rule_decimal, width=np.power(2,len(node_input_list)))]
    return(binary_rule)

def String_Rule_From_Binary(node_rule_binary,node_input_list):
    """Convert binary rule to BooleanNet format.

    Parameters
    ----------
    node_rule_binary : list of int
        Binary rule list corresponding to an output column of a truth table.
    node_input_list : list of str
        Variable names that correspond to each column of the truth table.

    Returns
    -------
    str
        BooleanNet representation of rule.

    """
    notConstant=False
    bit_previous=-1
    for i,bit in enumerate(node_rule_binary):
        if(i==0):
            bit_previous=bit
            continue
        if(bit!=bit_previous):
            notConstant=True
            break;

    if(not notConstant):
        rule_string=str(bit_previous)
    else:
        implicants_rule=[[int(binary) for binary in np.binary_repr(i,width=len(node_input_list))] for i,bit in enumerate(node_rule_binary) if bit==1]
        rule_string=" or ".join(
            ["("+" and ".join(
            [node_input_list[k] if(state==1) else "not "+node_input_list[k] for k,state in enumerate(implicant)]
        )+")"
            for implicant in implicants_rule]
        )
    return(rule_string)

def String_Rules_From_Binary(node_rules_binary_dictionary):
    """Convert from binary dictionary rule format to BooleanNet format.

    Parameters
    ----------
    node_rules_binary_dictionary : dictionary
        Binary dictionary representation of rules.

    Returns
    -------
    str
        BooleanNet representation of rules.

    """
    node_rules_string_dictionary={node_name:String_Rule_From_Binary(node_rule_input_list[0],node_rule_input_list[1])
                                  for node_name,node_rule_input_list in node_rules_binary_dictionary.items()}
    return(node_rules_string_dictionary)

def random_boolean_network_ensemble_kauffman(N,K,p,N_ensemble,seed=1000,write_boolean_network=False):
    """Generate a sample from the Kauffman NK RBN ensemble.

    Parameters
    ----------
    N : int
        Number of nodes of RBN.
    K : int
        Number of inputs of each node in the RBN.
    p : float
        Probability that each entry in each truth table output column is equal to 1.
    N_ensemble : int
        Number of networks to generate.
    seed : int
        Random seed for generating the RBN ensemble (the default is 1000).
    write_boolean_network : bool
        Whether to write each network in the ensemble as a CSV file in a new
        directory (the default is False).

    Returns
    -------
    RBN_ensemble_rules : list of str
        Each string are the Boolea rules of an ensemble in booleannet format.
        Each element in RBN_ensemble_rules can be used as an input for the
        format.booleannet2bnet function.

    """
    rd.seed(seed)
    random_boolean_type = "Kauffman NK"
    RBN_ensemble_rules=[]
    if(write_boolean_network):
        directory = "RBN"+"_N-"+str(N)+"_K-"+str(K)+"_p-"+str(p)+"_"+ datetime.now().strftime('%Y%m-%d%H-%M%S-') + str(uuid4())
        try:
            os.makedirs(directory)
        except FileExistsError:
            # directory already exists
            print ('Error: Creating directory ' +  directory)
            pass
    for n in range(N_ensemble):
        if(write_boolean_network):
            filename="RBN"+"_N-"+str(N)+"_K-"+str(K)+"_p-"+str(p)+"_Nensemble-"+str(n)+".csv"
            rbn=RandomBooleanNetworks()
            rbn.random_boolean_network(random_boolean_type,N,rbn_parameters=[K,p],filename=os.path.join(directory, filename))
            rules=rbn.random_boolean_network_Rules()
            RBN_ensemble_rules.append(rules)
        else:
            rbn=RandomBooleanNetworks()
            rbn.random_boolean_network(random_boolean_type,N,rbn_parameters=[K,p])
            rules=rbn.random_boolean_network_Rules()
            RBN_ensemble_rules.append(rules)
    return(RBN_ensemble_rules)

def get_criticality_K_Kauffman(p):
    """The Kauffman RBN is at criticality when K = 2/(p(1-p)).

    Parameters
    ----------
    p : float
        Probability that each entry in each truth table output column is equal to 1.

    Returns
    -------
    K_criticality : int
        Number of inputs of each node in the RBN.

    """


    K_criticality=[2.0/(p*(1-p))]
    return(K_criticality)

def get_criticality_p_Kauffman(K):
    """The Kauffman RBN is at criticality when K = 2/(p(1-p)).

    Parameters
    ----------
    K : int
        Number of inputs of each node in the RBN.


    Returns
    -------
    p_criticality : float
        Probability that each entry in each truth table output column is equal to 1.

    """
    p_criticality=[]
    if(K<2):
        p_criticality=[]
    elif(K==2):
        p_criticality=[0.5]
    elif(K>2):
        p_criticality=[0.5*(1+np.sqrt(1.0-2.0/K)),0.5*(1-np.sqrt(1.0-2.0/K))]
    return(p_criticality)



def main():
    import pystablemotifs as sm
    from sys import argv
    """
    Example Usage:
    python -m pystablemotifs <model file>

    here <model file> is the path to a model file, e.g. "./models/simple_model.txt".

    This will print the rules, a summary of the system's attractor repertoire, and
    trap space control interventions using default methods and parameters.
    """

    print("""
    This function is for demonstration. It runs the pystablemotifs attractor
    identification routine and identifies control interventions to target each
    identified attractor's trap space, provided it is nontrivial. Default function
    paramters are used. See files in the "Examples and Tutorials" directory for
    instructions regarding typical use cases.
    """)

    primes = sm.format.import_primes(argv[1],remove_constants=True)
    print("RULES")
    sm.format.pretty_print_prime_rules(primes)
    print()
    print("Analyzing network . . .")
    ar = sm.AttractorRepertoire.from_primes(primes,max_simulate_size=20)
    print("Analysis complete.")
    print()
    ar.summary()
    print("-"*10)

    print("\nComputing attractor control . . .")
    for attractor in ar.attractors:
        target = attractor.logically_fixed_nodes
        if len( target ) > 0 and attractor.guaranteed == True:
            print("-"*10)
            print("Target:",target)
            intervention = ar.reprogram_to_trap_spaces(target)
            print("Interventions:")
            for x in intervention: print(x)
    print("Complete.")

if __name__=='__main__':
    main()



import pystablemotifs.reduction as sm_reduction


class Attractor:
    """Stores attractor data for a reduced network. Automatically initialized by
    the AttractorRepertoire class.

    Parameters
    ----------
    reduction : reduction.MotifReduction
        Motif reduction to use as the representative (see attributes).
    reduction_attractor_id : int
        Reduction id to use for the representative (see attributes)

    Attributes
    ----------
    logically_fixed_nodes : partial state dictionary
        The nodes that are fixed by percolation on the expanded network (i.e.,
        not by up-stream oscillations)
    representative : reduction.MotifReduction, int tuple
        Entry 0 is a maximally reduced reduction.MotifReduction object that
        contains the attractor. In general, other such objects conain the
        attractor, but they will correspond to equivalent reduced networks.
        Entry 1 is a unique identifier number (integer) for the attractor within
        the reduced network; this is necessary in cases when a fully reduced n
        etwork contains multiple (complex) attractors.
    reductions : list of reduction.MotifReduction
        Maximally reduced MotifReductions that contain the attractor.
    attractor_dict : dictionary
        a dictionary describing the node states in the attractor according to
        the following key -
             1 variable is "ON"
             0 variable is "OFF"
             X variable is known to oscillate
             ? at least one such variable must oscillate
             ! the attractor may be false; if it is genuine, at least one such
               variable must oscillate
    stg : networkx.DiGraph
        The state transition graph corresponding to the attractor (if computed)
    fixed_nodes : partial state dictionary
        All node states that are known to be fixed in the attractor.
    oscillation_fixed_nodes : partial state dictionary
        Node states that are fixed in the attractor, but that are not fixed by
        percolation in the expanded network. These states are instead fixed by
        up-stream oscillation.
    reduced_primes : pyboolnet primes dictionary
        Update rules for the maximally reduced network that contains the attractor.
    n_unfixed : int
        Number of nodes that are not logically fixed.
    size_lower_bound : int
        Lower bound on number of states in attractor.
    size_upper_bound : int
        Upper bound on number of states in attractor.
    explored : bool
        True if all attractor states and transitions are explicitly computed.
    guaranteed : bool
        True if and only if the attractor is known to be genuine. If False, the
        attractor may not actually be stable.


    """

    def __init__(self, reduction, reduction_attractor_id):
        self.logically_fixed_nodes = reduction.logically_fixed_nodes
        self.representative = (reduction,reduction_attractor_id)
        self.reductions = [reduction]
        self.attractor_dict = reduction.attractor_dict_list[reduction_attractor_id]

        if reduction.no_motif_attractors is not None:
            self.stg = reduction.partial_STG.subgraph(reduction.no_motif_attractors[reduction_attractor_id])
        else:
            self.stg = None

        self.fixed_nodes = {k:v for k,v in self.attractor_dict.items() if v in [0,1]}

        # if reduction.fixed_rspace_nodes is not None:
        #     self.fixed_nodes.update(reduction.fixed_rspace_nodes)

        self.oscillation_fixed_nodes = {k:v for k,v in self.fixed_nodes.items() if k not in self.logically_fixed_nodes}

        if reduction.rspace_update_primes is not None:
            self.reduced_primes = reduction.rspace_update_primes
        else:
            self.reduced_primes = reduction.reduced_primes

        self.n_unfixed = len(self.attractor_dict) - len(self.fixed_nodes)

        if self.stg is not None:
            self.size_lower_bound = len(list(self.stg.nodes()))
            self.size_upper_bound = self.size_lower_bound
        else:
            if self.n_unfixed <= 1:
                self.size_lower_bound = self.n_unfixed + 1
                self.size_upper_bound = self.n_unfixed + 1
            else:
                self.size_lower_bound = 2
                self.size_upper_bound = 2**(self.n_unfixed)

        if '?' in self.attractor_dict.values():
            self.explored = False
            self.guaranteed = True
        elif '!' in self.attractor_dict.values():
            self.explored = False
            self.guaranteed = False
        else: # X's, 1's, 0's
            self.explored = True
            self.guaranteed = True

        self.reductions = [reduction]

    def add_reduction(self, reduction):
        """Add a reduction to the attractor. Does not check for compatibility.

        Parameters
        ----------
        reduction : reduction.MotifReduction
            Motif reduction that also contains the attractor.

        """
        self.reductions.append(reduction)




try:
    import pyboolnet
except ImportError as e:
    print(e)
    print("""
pyboolnet can be installed via
pip install git+https://github.com/hklarner/pyboolnet
or, for a specific version, via
pip install git+https://github.com/hklarner/pyboolnet@3.0.5
""")
    raise

from pystablemotifs import format
from pystablemotifs import drivers
from pystablemotifs import reduction
from pystablemotifs import time_reversal
from pystablemotifs import restrict_space
from pystablemotifs import succession
from pystablemotifs import random_boolean_networks
from pystablemotifs.Attractor import Attractor
from pystablemotifs.AttractorRepertoire import AttractorRepertoire
from pystablemotifs import export

__version__ = '3.0.3'



import pyboolnet.prime_implicants
import pyboolnet.boolean_logic
from pyboolnet.external.bnet2primes import bnet_text2primes
import re

# Convert rules from BooleanNet format to pyboolnet format
def booleannet2bnet(rules):
    """Converts BooleanNet rules to BNet format.
    e.g., an input of
    "A*=B or C and not D"
    returns
    A,  B | C & !D

    Also replaces ~ with !

    Parameters
    ----------
    rules : str
        BooleanNet formatted rules.

    Returns
    -------
    str
        BNET formatted rules.

    """
    s = re.sub("\s*\*\s*=\s*",",\t",rules) # replace "=" with ",\t"
    s = re.sub("\s+not\s+"," !",s, flags=re.IGNORECASE) # not -> !
    s = re.sub("\(\s*not\s+","(!",s, flags=re.IGNORECASE) # not -> ! (with parens)
    s = re.sub("\s*~\s*"," !",s, flags=re.IGNORECASE) # ~ -> !
    s = re.sub("\s+and\s+"," & ",s, flags=re.IGNORECASE) # and -> &
    s = re.sub("\s+or\s+"," | ",s, flags=re.IGNORECASE) # or -> |
    s = re.sub("False","0",s, flags=re.IGNORECASE) # False -> 0 (ignore case)
    s = re.sub("True","1",s, flags=re.IGNORECASE) # True -> 1 (ignore case)
    return s

# Convert rules from CellCollective format to pyboolnet format
def cellcollective2bnet(rules):
    """Converts CellCollective rules to BNet format.
    e.g., an input of
    "A = B OR C AND NOT D"
    returns
    A,  B | C & !D

    Also replaces ~ with !

    Parameters
    ----------
    rules : str
        CellCollective formatted rules.

    Returns
    -------
    str
        BNET formatted rules.

    """
    s = re.sub("\s*=\s*",",\t",rules) # replace "=" with ",\t"
    s = re.sub("\s+not\s+"," !",s, flags=re.IGNORECASE) # not -> ! (ignore case)
    s = re.sub("\(\s*not\s+","(!",s, flags=re.IGNORECASE) # not -> ! (with parens)
    s = re.sub("\s*~\s*"," !",s) # ~ -> !
    s = re.sub("\s+and\s+"," & ",s, flags=re.IGNORECASE) # and -> &
    s = re.sub("\s+or\s+"," | ",s, flags=re.IGNORECASE) # or -> |
    s = re.sub("False","0",s, flags=re.IGNORECASE) # False -> 0 (ignore case)
    s = re.sub("True","1",s, flags=re.IGNORECASE) # True -> 1 (ignore case)
    return s
def bnet2sympy(rule):
    """Converts a BNet string expression to a sympy string expression.

    Parameters
    ----------
    rule : str
        Boolean expression in BNET format.

    Returns
    -------
    str
        Expression in sympy format.

    """

    crule = re.sub("!","~",rule)
    crule = re.sub("[\b\(]1[\b\)]","(x | ~x)",crule)
    crule = re.sub("[\b\(]0[\b\)]","(x & ~x)",crule)
    crule = re.sub("True","(x | ~x)",crule)
    crule = re.sub("False","(x & ~x)",crule)

    return crule

def sympy2bnet(rule):
    """Converts a sympy string expression to a BNET string expression.

    Parameters
    ----------
    rule : str
        Boolean expression in sympy format.

    Returns
    -------
    str
        Expression in BNET format.

    """
    crule = re.sub("~","!",rule)
    crule = re.sub("True","1",crule)
    crule = re.sub("False","0",crule)
    return crule

def create_primes(rules,remove_constants = False):
    """Convert a BooleanNet or BNET string into a pyboolnet primes dictionary.

    Parameters
    ----------
    rules : str
        BooleanNet or BNET formatted rules. Hybrid formats are accepted as well.
        For the CellCollective format, use import_primes to read rules from the
        relevant files.
    remove_constants : bool
        Whether or not to remove and percolate constant input values (the default
        is False).

    Returns
    -------
    pyboolnet primes dictionary
        Update rules in pyboolnet format.

    """
    primes = bnet_text2primes(booleannet2bnet(rules))
    if remove_constants:
        pyboolnet.prime_implicants.percolate(primes,remove_constants=True,copy=False)
    return primes

def remove_comment_lines(stream, comment_char="#"):
    """Removes commented out lines from stream, e.g., those starting with '#'.

    Parameters
    ----------
    stream : iterable of str
        Lines from which comments should be excluded.
    comment_char : str
        Lines beginning with this character will be excluded.

    Returns
    -------
    list of str
        Lines that do not begin with comment_char.

    """
    lines = list(stream)
    lines = filter(lambda x: not x.startswith(comment_char), lines)
    rules = "".join(lines)
    return rules

def import_primes(fname, format='BooleanNet', remove_constants=False):
    """Import boolean rules from file and return pyboolnet formatted primes list.

    Parameters
    ----------
    fname : str
        Path to (plaintext) file containing Boolean rules in format specified
        by the 'format' option.
        Path to Boolean Expressions folder in case of CellCollective format.
    format : str
         Boolean rule format; options are 'BooleanNet' or 'BNet' or 'CellCollective'
         (the default is 'BooleanNet').
    remove_constants : bool
         If True, variables that are constant are removed and their influence is
         percolated. Otherwise, they remain and we consider initial conditions
         in opposition to their values (the default is False).

    Returns
    -------
    pyboolnet primes dictionary
        Update rules in pyboolnet format.

    """

    if format == 'CellCollective':
        rules1 = remove_comment_lines(open(fname + '/expr/expressions.ALL.txt'))
        external = remove_comment_lines(open(fname + '/expr/external_components.ALL.txt'))
        # In the case of CellCollective format, external components are given
        # rules of the form "A = A"
        lines = external.splitlines()
        for i in range(len(lines)):
            lines[i] = lines[i] + ' = ' + lines[i]
        rules2 = "\n".join(lines)
        rules = rules1 + "\n" + rules2
        rules = cellcollective2bnet(rules)

    else:
        rules = remove_comment_lines(open(fname))
        if format == 'BooleanNet':
            rules = booleannet2bnet(rules)
        elif format == 'BNet':
            rules = rules
        else:
            raise ValueError('Unrecognized format',format)

    primes = bnet_text2primes(rules)
    if remove_constants:
        pyboolnet.prime_implicants.percolate(primes,remove_constants=True,copy=False)
    return primes


def _bnetDNF2list(bnet):
    """Converts a BNet string expression to a list of prime implicant dictionaries.
    Requires that the input be in disjunctive normal form, but this is not checked
    explicitly. This is used as a helper function during algebraic simplification.

    Parameters
    ----------
    bnet : str
        BNET formatted expression in disjunctive normal form.

    Returns
    -------
    list of partial state dictionaries
        Variable states specified by each dictionary are to be thought of as "AND"-
        connected, and the dictionaries as "OR"-connected.

    """

    if bnet == "0":
        return []
    elif bnet == "1":
        return [{}]

    bnetList = []
    bnet_trim = re.sub("[\s\(\)]*","",bnet) # remove all whitespace and parens
    LL = [x.split('&') for x in bnet_trim.split('|')]

    for L in LL:
        Ldict = {}
        contradiction = False
        for literal in L:
            if literal[0]=="!":
                n = literal[1:]
                s = 0
            else:
                n = literal
                s = 1
            if n not in Ldict:
                Ldict[n] = s
            elif Ldict[n] != s:
                contradiction = True
                break
        if not contradiction:
            bnetList.append(Ldict)
    return bnetList

def _build_rule_using_bnet_dnfs(expr0,expr1):
    """Converts a BNet string expression (expr1) and its negation (expr0) to
    a pyboolnet rule list. Note that this function does not test for consistency
    between expr0 and expr1. This is used as a helper function during network reduction.

    Parameters
    ----------
    expr0 : str
        Rule, in BNET format, for the "OFF" state of a variable.
    expr1 : str
        Rule, in BNET format, for the "ON" state of a variable.

    Returns
    -------
    pyboolnet rule list
        The complementary expressions as they would appear in a pyboolnet primes
        dictionary for a variable whose update rule is given by expr1.

    """
    return [_bnetDNF2list(expr0),_bnetDNF2list(expr1)]

def statelist2dict(names,statestrings):
    """Converts a collection of statestrings to a dictionary.

    Parameters
    ----------
    names : list of str
        An ordered list of variable names; (alphabetical order is pyboolnet's
        default, e.g. sorted(primes)).
    c : iterable of str
        Each element should be a binary string, with each position corresponding
        to the variable name at the same position in names.

    Returns
    -------
    dictionary
        Dictionary summarizing c. If a node takes the same value in every state,
        the corresponding dictionary value matches its fixed value; otherwise,
        the dictionary value is 'X'.

    """
    d = {}
    for i,n in enumerate(names):
        for cs in statestrings:
            if n not in d:
                d[n] = cs[i]
                continue
            if cs[i] != d[n]:
                d[n] = 'X'
                break
    return d

def statestring2dict(statestring,names):
    """Converts a state string, which specifies a node in an STG, to the
    corresponding dictionary representation.

    Parameters
    ----------
    statestring : str
        A binary string, e.g., '01101'.
    names : list of str
        An ordered list of variable names; (alphabetical order is pyboolnet's
        default, e.g. sorted(primes)).

    Returns
    -------
    dictionary
        The keys are the elements of names and the values are the corresponding
        value in statestring.

    """
    sd = {}
    for i,c in enumerate(statestring):
        sd[names[i]]=int(c)
    return sd

def statedict2str(statedict):
    """Converts a state dictionary to a statestring using alphabetical sorting.

    Parameters
    ----------
    statedict : partial state dictionary
        State to convert to a binary string representation.

    Returns
    -------
    str
        A binary string, with each position corresponding
        to the variable name at the same position in the alphabetized keys in
        statedict.

    """
    return ''.join([str(statedict[x]) for x in sorted(statedict)])

def implicant2bnet(partial_state):
    """Converts a partial state dictionary to a BNet string
    e.g., {'A':1,'B':0} returns 'A & !B'

    Parameters
    ----------
    partial_state : partial state dictionary
        Partial state to convert.

    Returns
    -------
    str
        BNET representation of the partial state.

    """
    return ' & '.join(["!"+k for k in partial_state if not partial_state[k]]+[k for k in partial_state if partial_state[k]])

def rule2bnet(rule):
    """Converts a pyboolnet prime rule into a BNet string.
    e.g., [{'A':1,'B':0},{'C':0}] returns 'A & !B | !C'

    Parameters
    ----------
    rule : list of pyboolnet partial states
        Update rule to convert.

    Returns
    -------
    str
        BNET representation of Boolean expression.

    """
    # There are two special cases:
    # [] is identically 0
    # [{}] is identically 1
    if rule == []: return '0'
    elif rule == [{}]: return '1'
    else: return ' | '.join([implicant2bnet(imp) for imp in rule])

def primes2bnet(primes):
    """A simpler version of pyboolnet's file_exchange.primes2bnet function with
    fewer options and less organized output. Should handle prime rules with
    tautologies better than the pyboolnet version though.

    Parameters
    ----------
    primes : pyboolnet primes dictionary
        Update rules to convert.

    Returns
    -------
    str
        BNET representation of update rules.

    """
    lines = []
    width = max([len(x) for x in primes]) + 3

    for name in primes:
        if primes[name][0] == [] or primes[name][1] == [{}]:
            expression = '1'
        elif primes[name][1] == [] or primes[name][0] == [{}]:
            expression = '0'
        else:
            expression = ' | '.join(['&'.join([x if term[x]==1 else '!'+x for x in term]) for term in primes[name][1]  ])

        lines+= [(name+',').ljust(width)+expression]
    lines+=['']

    return "\n".join(lines)

def primes2booleannet(primes, header=""):
    """Convert a pyboolnet primes dictionary to a BooleanNet string reperesentation.

    Parameters
    ----------
    primes : pyboolnet primes dictionary
        Update rules to convert.
    header : str
        Text to include at the beginning of the file, e.g., comment lines. For
        example, the legacy Java version of StableMotifs requires rules files to
        begin with the line "#BOOLEAN RULES".

    Returns
    -------
    str
        BooleanNet representation of update rules.

    """
    lines = []
    width = max([len(x) for x in primes]) + 3

    for name in primes:
        if primes[name][0] == [] or primes[name][1] == [{}]:
            expression = '1'
        elif primes[name][1] == [] or primes[name][0] == [{}]:
            expression = '0'
        else:
            expression = ' or '.join([' and '.join([x if term[x]==1 else 'not '+x for x in term]) for term in primes[name][1]  ])

        lines+= [(name+'*=').ljust(width)+expression]
    lines+=['']

    return header + "\n".join(lines)

def pretty_print_primes(primes):
    """Prints pyboolnet a prime dictionary in a more readable format. Prints both
    state updates (1 and 0).

    Parameters
    ----------
    primes : pyboolnet primes dictionary
        Update rules to print.

    """
    for k,v in primes.items():
        for p in v[0]: print(p,"=> !"+k)
        for p in v[1]: print(p,"=>  "+k)

def pretty_print_prime_rules(primes):
    """Prints pyboolnet a prime dictionary as Boolean rules
    The output format is of the form:
    A* = B & C | !D, for example.

    Parameters
    ----------
    primes : pyboolnet primes dictionary
        Update rules to print.

    """

    if primes is None:
        return ""

    for k,v in primes.items():
        s = k + "* = "
        sl = []
        for c in v[1]:
            sll = []
            for kk,vv in c.items():
                if vv: sli = kk
                else: sli = '!'+kk
                sll.append(sli)
            if len(sll) > 0:
                sl.append(' & '.join(sll))
        if len(sl) > 0:
            s += ' | '.join(sl)
        if v[1]==[]:
            s = k + "* = 0"
        if v[1]==[{}]:
            s = k + "* = 1"
        print(s)

def pretty_print_rspace(L,simplify=True,silent=True):
    """Produces string representation of the Boolean rule describing the input
    rspace L (see restrict_space.rspace).

    Parameters
    ----------
    L : rspace list
        Restrict space list (see restrict_space.rspace for details).
    simplify : bool
        Whether to simplify the rule (the default is True).
    silent : bool
        Whether to supress output of the rule (the default is True).

    Returns
    -------
    str
        BNET expression that is true in and only in the rspace specified by L.

    """
    L_trim = [x for x in L if x != [{}]]
    u = []
    for x in L_trim:
        t = []
        for y in x:
            s = []
            for k,v in y.items():
                if v: s.append(k)
                else: s.append('!'+k)
            t.append(' & '.join(s))
        if len(t) > 0: u.append('( '+' | '.join(t)+' )')
    s=' & '.join(u)
    if simplify:
        s = pyboolnet.boolean_logic.minimize_espresso(s)
    if not silent:
        print(s)

    return s



import networkx as nx
import itertools as it
import matplotlib
import matplotlib.pyplot as plt

from copy import deepcopy

import pystablemotifs.reduction as sm_reduction
import pystablemotifs.format as sm_format
import pystablemotifs.drivers as sm_doi

class SuccessionDiagram:
    """Class describing the succession diagram of a Boolean system. See, e.g.,
    Zanudo and Albert (2015) or Rozum et al. (2021).

    Attributes
    ----------
    motif_reduction_dict : dictionary
        MotifReduction-valued dictionary with integer (index) keys (see reduction.py).
    digraph : networkx.DiGraph
        Topological structure of hte succession diagram. Nodes are integers that
        align with the enteries of motif_reduction_dict.

    """

    def __init__(self):
        self.motif_reduction_dict = {}
        self.digraph = nx.DiGraph()

    def get_motifs(self):
        """Extract the stable motifs of a system and its reduced networks from its
        attractor repertoire. Notably, these include both the system's primary stable
        motifs and conditionally stable motifs (see, e.g., Deritei et al. 2019).

        Returns
        -------
        list of dictionaries
            Stable motifs that appear in the system or during reduction (in no
            particular order).

        """
        SM_set = set([])
        for reduction in self.motif_reduction_dict.values():
            if len(reduction.motif_history)>0:
                SM_set.add(frozenset(reduction.motif_history[-1].items()))
        return [dict(s) for s in SM_set]

    def find_motif_permutation(self,motif_history):
        """Check whether some permutation of the input motif_history is already
        represented in the succession diagram. If so, return the preexisting
        reduction's index and the permutation that maps between the two histories.

        Parameters
        ----------
        motif_history : list of partial state dictionaries
            Stable motifs that can lock in to give a given reduced network (in
            order).

        Returns
        -------
        reduction_index : int
            Index of the preexisting reduced network. This value is None if no such
            reduced network exists.
        permutation : list of int
            Permutation that maps the preexisting history to the input history.
            This value is None if no such history exists.

        """
        for reduction_index,mr in self.motif_reduction_dict.items():
            if len(mr.motif_history) == len(motif_history):
                if all([x in mr.motif_history for x in motif_history]):
                    permutation = []
                    for x in motif_history:
                        permutation.append(mr.motif_history.index(x))
                    return reduction_index,permutation # We already have an equivalent motif in the list
        return None,None

    def add_motif_permutation(self,reduction_index,permutation):
        """Adds a permutation of a preexisting stable motif history to a precomputed
        MotifReduction object.

        Parameters
        ----------
        reduction_index : int
            Index of the preexisting reduced network.
        permutation : list of int
            Permutation that maps the preexisting history to the input history.

        """
        self.motif_reduction_dict[reduction_index].merged_history_permutations.append(permutation)
        for child in nx.topological_sort(self.digraph):
            for parent in self.digraph.predecessors(child):
                for parent_perm,child_perm in it.product(
                    self.motif_reduction_dict[parent].merged_history_permutations,
                    self.motif_reduction_dict[child].merged_history_permutations):
                    new_perm = child_perm.copy()
                    for i,p in enumerate(parent_perm):
                        new_perm[i] = child_perm[p]
                    if not new_perm in self.motif_reduction_dict[child].merged_history_permutations:
                        self.motif_reduction_dict[child].merged_history_permutations.append(new_perm)

    def find_equivalent_reduction(self,fixed):
        """Extracts the MotifReduction object that has the frozen node values
        specified by fixed, if such an object exists (returns None otherwise).

        Parameters
        ----------
        fixed : partial state dictionary
            Nodes values that have been fixed and reduced by stable motifs and their
            logical domain of influence.

        Returns
        -------
        MotifReduction
            Reduced network that has the frozen node values specified by fixed,
            if such an object exists (returns None otherwise).

        """
        for reduction in self.motif_reduction_dict.values():
            if reduction.logically_fixed_nodes == fixed:
                return reduction
        return None

    def add_motif_reduction(self,motif_reduction):
        """Inserts a given MotifReduction into the succession diagram. Does not
        check for consistency, but will insert a properly constructed MotifReduction
        into the correct place in the diagram.

        Parameters
        ----------
        motif_reduction : MotifReduction
            Reduced network to be appended to the succession diagram.

        """
        if self.motif_reduction_dict == {}:
            self.unreduced_primes = motif_reduction.reduced_primes

        # note: N is computed BEFORE the new reduction is added,
        # so it will be the reduction index AFTER the reduction is added.
        N = len(self.motif_reduction_dict)
        new_set = set([frozenset(tuple(x.items())) for x in motif_reduction.motif_history])
        if N == 0:
            self.digraph.add_node(0)
        else:
            for i,reduction in self.motif_reduction_dict.items():
                old_set = set([frozenset(tuple(x.items())) for x in reduction.motif_history])

                # see if we're adding a parent of an existing reduction
                if len(old_set) == len(new_set) + 1:
                    diff_set = old_set - new_set
                    if len(diff_set) == 1:
                        missing_motif = dict(diff_set.pop())
                        if missing_motif in motif_reduction.stable_motifs:
                            self.digraph.add_edge(N,i)
                # see if we're adding a child of an existing reduction
                elif len(old_set) == len(new_set) - 1:
                    diff_set = new_set - old_set
                    if len(diff_set) == 1:
                        missing_motif = dict(diff_set.pop())
                        if missing_motif in reduction.stable_motifs or len(new_set) == 1: # valid motif OR source combination
                            self.digraph.add_edge(i,N)

        self.motif_reduction_dict[N] = motif_reduction
        self.add_motif_permutation(N,list(range(len(motif_reduction.motif_history))))

    def reductions_indices_with_states(self,logically_fixed,optimize=True):
        """Find all reductions (by index) that have the nodes states specified
        logically fixed.

        Parameters
        ----------
        logically_fixed : partial state dictionary
            Nodes states that should be fixed in all returned network reductions.
        optimize : bool
            Whether to remove reduced networks that are subnetworks of valid
            reductions. This is generally recommended so as to obtain the most
            parsimonious control strategies (the default is True).

        Returns
        -------
        list of int
            Indices of reduced networks that have the appropriate fixed states.

        """
        if not optimize:
            # NOTE: This finds all reductions, not just those closest to the root
            target_indices = []
            for i,reduction in self.motif_reduction_dict.items():
                if logically_fixed.items() <= reduction.logically_fixed_nodes.items():
                    target_indices.append(i)
            return target_indices
        else:
            # The unoptimized result
            target_indices_unoptimized = self.reductions_indices_with_states(logically_fixed,optimize=False)
            target_indices_all = target_indices_unoptimized.copy()

            # Add nodes that inevitably reach nodes in the unoptimized result
            nodes_to_consider = [x for x in self.digraph if (not x in target_indices_all
                and self.motif_reduction_dict[x].terminal == "no")]
            for i in nodes_to_consider:
                bad_sinks = [x for x in nx.descendants(self.digraph,i) | set([i]) if (
                    not self.motif_reduction_dict[x].terminal == "no" # is a sink and . . .
                    and not x in target_indices_unoptimized)] # does not have the right nodes fixed

                if len(bad_sinks) == 0:
                    target_indices_all.append(i)

            # remove nodes who have ancestors that would work instead, as those
            # will be closer to the root and easier to control
            target_indices = []
            for target_index in target_indices_all:
                if set(nx.ancestors(self.digraph,target_index)) & set(target_indices_all) == set():
                    target_indices.append(target_index)

            return target_indices

    def reduction_drivers(self,target_index,method='internal',max_drivers=None,GRASP_iterations=None):
        """Find control strategies that lead to the reduced network specified by
        the target index. Several control strategies are implemented. See
        succession.SuccessionDiagram.reprogram_to_trap_spaces for a detailed
        description of control methods available. Generally, this method should
        not be used directly. Instead, use reprogram_to_trap_spaces.

        Parameters
        ----------
        target_index : int
            Index of the target reduced network.
        method : str
            One of 'internal', 'minimal', or 'GRASP'. See
            succession.SuccessionDiagram.reprogram_to_trap_spaces for details.
        max_drivers : int
            Maximum number of driver nodes to consider (not used in GRASP methods).
            If none, the upper limit is given by the number of free variables
            (the default is None).
        GRASP_iterations : int
            Number of times to construct GRASP driver sets; only used in GRASP
            methods. If none, the number of iterations is chosen based on the
            network size (the default is None).

        Returns
        -------
        list
            Control strategies found; interpretation depends on method selected
            See succession.SuccessionDiagram.reprogram_to_trap_spaces for details.

        """
        methods = ['internal','minimal','GRASP']
        assert method in methods, ' '.join(["method argument of reduction_drivers must be among",str(methods)])
        drivers = []
        for path in nx.all_simple_paths(self.digraph,0,target_index):
            path_motif_history=[]
            path_drivers = []
            for ind in path:
                if ind == 0:
                    ind_prev = ind
                    continue
                path_motif_history += [x for x in self.motif_reduction_dict[ind].motif_history if not x in path_motif_history]

                if method == 'internal':
                    history_drivers = sm_doi.internal_drivers(path_motif_history[-1],
                        self.motif_reduction_dict[ind_prev].reduced_primes,
                        max_drivers=max_drivers)
                elif method == 'GRASP':
                    history_drivers = sm_doi.GRASP(path_motif_history[-1],
                        self.motif_reduction_dict[ind_prev].reduced_primes,
                        GRASP_iterations = GRASP_iterations)
                    if len(history_drivers) == 0:
                        history_drivers = [path_motif_history[-1]]
                elif method == 'minimal':
                    history_drivers = sm_doi.minimal_drivers(path_motif_history[-1],
                        self.motif_reduction_dict[ind_prev].reduced_primes,
                        max_drivers=max_drivers)

                path_drivers.append(history_drivers)

                ind_prev = ind
            if method == 'internal':
                # merge control sets along the path
                for control_sequence in it.product(*path_drivers):
                    control_set = {k:v for x in control_sequence for k,v in x.items()}
                    if not control_set in drivers:
                        drivers.append(control_set)
            elif method == 'minimal' or method == 'GRASP':
                drivers.append(path_drivers)
        return drivers

    def reprogram_to_trap_spaces(self,logically_fixed,target_method='history',driver_method='internal',max_drivers=None,GRASP_iterations=None,GRASP_score_override=None):
        """Find driver sets that lead to fixing the node states specified.

        Parameters
        ----------
        logically_fixed : partial state dictionary
            Targeted fixed nodes.
        target_method : str
            Either 'history' or 'merge'; see Notes below for details.
        driver_method : str
            Either 'internal', 'minimal', or 'GRASP' see Notes below for details.
        max_drivers : int
            Maximum number of driver nodes to consider (not used in GRASP methods).
            If none, the upper limit is given by the number of free variables
            (the default is None).
        GRASP_iterations : int
            Number of times to construct GRASP driver sets; only used in GRASP
            methods. If none, the number of iterations is chosen based on the
            network size (the default is None).
        GRASP_score_override : function
            Optional heuristic score function override (see drivers.GRASP
            for details). Only used in GRASP methods (the default is None).

        Returns
        -------
        list
            Control strategies found; interpretation depends on method selected
            See Notes below for details.

        Notes
        -----
        The various combinations of target_method and driver_method options result
        in different control strategies, which are outlined below.

        target_method = history, driver_method = internal:
        Finds all shortest stable motif histories that result in the target node states
        being logically fixed. Each stable motif is searched for internal driver nodes.
        The resulting internal drivers are combined into a single  control set. The
        return value consists of all such control sets for all  stable motif histories.
        Each control set eventually becomes self-sustaining.

        target_method = history, driver_method = minimal:
        Similar to the history method, except the search for stable motif drivers
        includes external driver nodes for the motif and does not extend to driver sets
        of larger size once one driver set has been found for a motif. Because the
        search includes external driver nodes, special care must be taken in interpreting
        the effect of the drivers, as their influence may impact the effect of motifs
        stabilizing. Thus, the control is only guaranteed to work if the interventions
        are temporary and implemented in the order specified by the motif history.

        For this reason, the output consists of lists of ordered interventions.
        Each element of the return value is a list of lists of dictionaries. Each
        element of the return value represents a control strategy. To implement such
        a strategy, select a dictionary from the first element of the strategy and
        fix the node states it specifies until their influence has propagated through
        the system. Then repeat this process iteratively for each element of the strategy
        list, in order. For example, if
        nonredundant_drivers = [ [[{'xD':1,'xE=1'}]], [[{'xA':1},{'xB':1}],[{'xC':1}]] ]
        then there are two control strategies available:
        1) fix xD=xE=1 temporarily and
        2) first fix either xA=1 or xB=1 temporarily, then fix xC=1 temporarily.

        target_method = history, driver_method = GRASP:
        The same as history, minimal, except external driver nodes are searched for
        using the GRASP algorithm using GRASP_iterations iterations.

        target_method = merge, driver_method = internal:
        Finds all shortest stable motif histories that result in the target node states
        being logically fixed. All node states in the motifs in the history are merged
        into a stable module dictionary. This is then searched for internal driver
        nodes. Each element of the return value is a dictionary corresponding to a
        control set. Each control set eventually becomes self-sustaining.

        target_method = merge, driver_method = minimal:
        Similar to the merge method, except the search for drivers is conducted over
        all nodes, not just those internal to the merged stable module. Furthermore,
        the search is truncated when a control set is found such that the search does
        not proceed to driver sets larger than the smallest found. Each element of
        the return value is a dictionary corresponding to a control set. The control
        sets are only guaranteed to result in activation of the target if they are
        temporary interventions.

        target_method = merge, driver_method = GRASP:
        The same as merge, minimal, except external driver nodes are searched for
        using the GRASP algorithm using GRASP_iterations iterations.

        """

        #methods = ['history','merge','minimal_history','minimal_merge']
        target_methods = ['history','merge']
        driver_methods= ['internal','minimal','GRASP']
        assert target_method in target_methods, ' '.join(["target_method argument of reprogram_to_trap_spaces must be among",str(target_methods)])
        assert driver_method in driver_methods, ' '.join(["driver_method argument of reprogram_to_trap_spaces must be among",str(driver_methods)])
        drivers = []

        if driver_method == 'GRASP' and GRASP_iterations is None:
            if target_method == 'merge':
                GRASP_iterations = len(self.unreduced_primes)**2
            if target_method == 'history':
                GRASP_iterations = 2*len(self.unreduced_primes)

        target_indices = self.reductions_indices_with_states(logically_fixed)

        if target_method == 'history':
            for target_index in target_indices:
                drivers += self.reduction_drivers(target_index,max_drivers=max_drivers,GRASP_iterations=GRASP_iterations,method=driver_method)
        elif target_method == 'merge':
            for target_index in target_indices:
                target_history = self.motif_reduction_dict[target_index].motif_history
                motif_merger = {k:v for d in target_history for k,v in d.items()}

                # Because we are potentially dealing with external drivers, we
                # want to make sure the external drivers do not interfere with
                # the motif's ability to drive downstream targets
                # if driver_method != 'internal':
                #     motif_merger.update(logically_fixed)

                if driver_method == 'GRASP':
                    merger_drivers = sm_doi.GRASP(motif_merger,self.unreduced_primes,GRASP_iterations)
                    if len(merger_drivers) == 0:
                        merger_drivers = [motif_merger.copy()]
                elif driver_method == 'minimal':
                    merger_drivers = sm_doi.minimal_drivers(motif_merger,
                        self.unreduced_primes,max_drivers=max_drivers)
                elif driver_method == 'internal':
                    merger_drivers = sm_doi.internal_drivers(motif_merger,
                        self.unreduced_primes,max_drivers=max_drivers)

                drivers += [x for x in merger_drivers if not x in drivers]

        drivers = sorted(drivers,key=lambda x: len(x))

        # Next, we remove redundant control sets.
        # In the minimal_history scheme, a set x is redundant if there is a y
        # that contains all control sets of x in the same order.
        # In all other schemes, x is redundant if it is a subset of some y.
        nonredundant_drivers = []
        if target_method == 'history' and (driver_method == 'minimal' or driver_method == 'GRASP'):
            nonredundant_drivers = []
            for x in drivers: # drivers is sorted from fewest timesteps to most
                add_x = True
                for y in nonredundant_drivers: # note: len(y) <= len(x)
                    for offset in range(len(x)-len(y)+1):
                        for i in range(len(y)):
                            # want to check that x[i+offset] is a special case of y[i]
                            for j in range(len(y)):
                                xset = set([frozenset(tuple(xdict.items())) for xdict in x[i+offset]])
                                yset = set([frozenset(tuple(xdict.items())) for xdict in y[i]])
                                if yset <= xset:
                                    add_x = False
                                    break
                            if not add_x: break
                        if not add_x: break
                    if not add_x: break
                if add_x:
                    nonredundant_drivers.append(x)

        else:
            for i in range(len(drivers)):
                use_i = True
                for j in range(i): # i > j
                    if drivers[i].items() >= drivers[j].items():
                        use_i = False
                        break
                if use_i: nonredundant_drivers.append(drivers[i])

        return nonredundant_drivers

def build_succession_diagram(primes, fixed=None, motif_history=None, diagram=None, merge_equivalent_motifs=True,max_simulate_size=20,prioritize_source_motifs=True,max_stable_motifs=10000, MPBN_update=False):
    """Recursively construct a succession diagram from the input update rules.
    Generally, it is preferable to construct this from within the AttractorRepertoire
    class (using, e.g., AttractorRepertoire.from_primes).

    Parameters
    ----------
    primes : pyboolnet primes dictionary
        Update rules.
    fixed : partial state dictionary
        Used only for recursion. Specifies nodes to be fixed in the next reduced
        network to be added to the diagram.
    motif_history : list of partial state dictionaries
        Used only for recursion. Specifies stable motif history for the next reduced
        network to be added to the diagram.
    diagram : SuccessionDiagram
        Used only for recursion. The SuccessionDiagram object that is under
        construction.
    merge_equivalent_motifs : bool
        If False, equivalent reduced networks have their data recomputed and
        copied. Making this False is only recommended if the succession diagram
        must be represented in a form that has no feedforward loops; making this
        True provides large computational advantages, both in terms of speed and
        memory usage (the default is True).
    max_simulate_size : int
        Maximum number of variables for which to brute-force build a state
        transition graph (the default is 20).
    prioritize_source_motifs : bool
        Whether source nodes should be considered first (the default is True).
    max_stable_motifs : int
        Maximum number of output lines for pyboolnet to process from the
        AspSolver (the default is 10000).
    MPBN_update : bool
        Whether MBPN update is used instead of general asynchronous update
        (the default is False).

    Returns
    -------
    SuccessionDiagram
        The succession diagram for the input update rules.

    """

    if diagram is None:
        diagram = SuccessionDiagram()

    if fixed is None:
        fixed = {}

    myMotifReductionToCopy = diagram.find_equivalent_reduction(fixed)

    if myMotifReductionToCopy is None:
        myMotifReduction=sm_reduction.MotifReduction(motif_history,fixed.copy(),primes,max_simulate_size=max_simulate_size,prioritize_source_motifs=prioritize_source_motifs,max_stable_motifs=max_stable_motifs,MPBN_update=MPBN_update)
    else:
        myMotifReduction = deepcopy(myMotifReductionToCopy)
        myMotifReduction.motif_history = motif_history.copy()
        myMotifReduction.merged_history_permutations = []

    diagram.add_motif_reduction(myMotifReduction)

    # Prioritize source nodes
    if myMotifReduction.merged_source_motifs is None:
        stable_motif_list = myMotifReduction.stable_motifs
    else:
        stable_motif_list = myMotifReduction.merged_source_motifs

    for sm in stable_motif_list:
        if merge_equivalent_motifs:
            perm_index,perm = diagram.find_motif_permutation(myMotifReduction.motif_history+[sm])
            if not perm_index is None:
                diagram.add_motif_permutation(perm_index,perm)
        if not merge_equivalent_motifs or perm_index is None:
            np,fixed2 = sm_reduction.reduce_primes(sm,primes)
            fixed3 = fixed.copy()
            fixed3.update(fixed2)
            diagram = build_succession_diagram(np,fixed3,myMotifReduction.motif_history+[sm],
                diagram, merge_equivalent_motifs=merge_equivalent_motifs,
                max_simulate_size=max_simulate_size,
                max_stable_motifs=max_stable_motifs,
                prioritize_source_motifs=prioritize_source_motifs,
                MPBN_update=MPBN_update)
    return diagram



def time_reverse_primes(primes):
    """Computes the time reversal of the input system (under general asynchronous update).
    The time reverse system has the same STG as the input system, but with each edge reversed.

    Parameters
    ----------
    primes : pyboolnet prime dictionary
        System update rules.

    Returns
    -------
    trprimes : pyboolnet prime dictionary
        Time-reversed system update rules.

    """
    trprimes = {}
    for k,v in primes.items():
        trv = [[],[]]
        for i in [0,1]:
            for p in v[i]:
                pc = p.copy()
                if k in pc: pc[k] = int(not pc[k])
                trv[int(not i)].append(pc)
        trprimes[k]=trv.copy()
    return trprimes



import pyboolnet.trap_spaces
import pyboolnet.prime_implicants
import pyboolnet.interaction_graphs
import pyboolnet.digraphs
from pyboolnet.external.bnet2primes import bnet_text2primes
import itertools as it
import networkx as nx
import re
import sympy

import pystablemotifs.time_reversal as sm_time
import pystablemotifs.restrict_space as sm_rspace
import pystablemotifs.format as sm_format
import pystablemotifs.drivers as sm_doi

def simplify_primes(primes):
    """Simplifies pyboolnet primes (e.g., A | A & B becomes A)

    Parameters
    ----------
    primes : pyboolnet primes dictionary
        Rules to simplify.

    Returns
    -------
    pyboolnet primes dictionary
        Simplified rules.

    """

    # reimport to force simplification
    if len(primes) > 0:
        return bnet_text2primes(sm_format.primes2bnet(primes))
    else:
        return primes

def reduce_primes(fixed,primes):
    """Simplifies boolean rules when some nodes are held fixed

    Parameters
    ----------
    fixed : partial state dictionary
        Node states to be held fixed.
    primes : pyboolnet primes dictionary
        Update rules.

    Returns
    -------
    reduced_primes : pyboolnet primes dictionary
        Simplified update rules
    percolated_states : partial state dictionary
        Fixed node states (including inputs) that were simplified and removed.

    """
    reduced_primes = pyboolnet.prime_implicants.percolate(primes, add_constants=fixed, remove_constants=False, copy=True)
    percolated_states = pyboolnet.prime_implicants.find_constants(reduced_primes)
    pyboolnet.prime_implicants.remove_all_constants(reduced_primes, copy=False) # remove constants in place

    return simplify_primes(reduced_primes), percolated_states

def delete_node(primes, node):
    """Reduces Boolean rules given by primes by deleting the variable specified by
    node. The deleted node may not appear in its own update function. Any update
    rules depending on the deleted node will have that dependence replaced by
    the update function of the deleted node. The rules are simplified after node
    deletion.

    Parameters
    ----------
    primes : pyboolnet primes dictionary
        Update rules.
    node : str
        Name of the node to delete.

    Returns
    -------
    new_primes : pyboolnet primes dictionary
        The reduced primes.
    constants : partial state dictionary
        Node states that became logically fixed during simplification.


    """
    succ = set()
    for p in primes:
        p_is_succ = False
        for v in [0,1]:
            for term in primes[p][v]:
                if node in term:
                    succ.add(p)
                    p_is_succ = True
                    break
            if p_is_succ:
                break


    # uncomment for additional (and hopefully unnecessary), but slow safety checks.
    #G = pyboolnet.interaction_graphs.primes2igraph(primes)
    #assert not G.has_edge(node,node), ' '.join(["Node",str(node),"has a self-loop and cannot be deleted."])

    assert not node in succ, ' '.join(["Node",str(node),"has a self-loop and cannot be deleted."])
    new_primes = {k:v for k,v in primes.items() if not k == node}
    constants = {}

    neg = "!"+node

    expr0 = sm_format.rule2bnet(primes[node][0])
    expr1 = sm_format.rule2bnet(primes[node][1])

    for child in succ:#G.successors(node):
        # If we have already simplified this child node, skip it
        if child in constants:
            continue
        crule1 = sm_format.rule2bnet(primes[child][1])
        crule1 = simplify_using_expression_and_negation(node,expr0,expr1,crule1)
        crule0 = sm_format.rule2bnet(primes[child][0])
        crule0 = simplify_using_expression_and_negation(node,expr0,expr1,crule0)
        new_primes[child] = sm_format._build_rule_using_bnet_dnfs(crule0,crule1)

        pyboolnet.prime_implicants.percolate(new_primes, remove_constants=False, copy=False)
        constants.update(pyboolnet.prime_implicants.find_constants(new_primes))
        #pyboolnet.prime_implicants.remove_all_constants(new_primes, copy=False)

    pyboolnet.prime_implicants.percolate(new_primes, remove_constants=False, copy=False)
    constants.update(pyboolnet.prime_implicants.find_constants(new_primes))
    pyboolnet.prime_implicants.remove_all_constants(new_primes, copy=False)
    new_primes = simplify_primes(new_primes)
    return new_primes, constants

def simplify_using_expression_and_negation(node,expr0,expr1,bnet):
    """Simplify the expression bnet by substituting the value for node given by
    node = expr1 = !expr0 (does not check that expr1=!expr0)

    Parameters
    ----------
    node : str
        Name of node to substitute
    expr0 : str
        Expression to substitute for !node
    expr1 : str
        Expression to substitute for node
    bnet : str
        BNET expression in which to perform the substitutions.

    Returns
    -------
    str
        Simplified BNET expression after substitutions are performed.

    """

    neg = "!"+node
    crule = re.sub(rf'\b{neg}\b',"("+expr0+")",bnet)
    crule = re.sub(rf'\b{node}\b',"("+expr1+")",crule)
    crule = sm_format.bnet2sympy(crule)
    crule = str(sympy.to_dnf(sympy.simplify(sympy.sympify(crule))))
    crule = sm_format.sympy2bnet(crule)
    return crule

def remove_outdag(primes):
    """Removes the terminal directed acyclic part of the regulatory network. This
    part of the network does not influence the attractor repertoire.

    Parameters
    ----------
    primes : pyboolnet primes dictionary
        Update rules.

    Returns
    -------
    reduced : pyboolnet primes dictionary
        The reduced primes.
    constants : partial state dictionary
        Node states that became logically fixed during reduction.

    """
    if len(primes) == 0:
        return primes, {}
    G = pyboolnet.interaction_graphs.primes2igraph(primes)
    od = pyboolnet.digraphs.find_outdag(G)
    reduced = primes.copy()
    constants = {}
    for node in od:
        if node in reduced:
            reduced, nc = delete_node(reduced, node)
            constants.update(nc)
    return reduced, constants

def deletion_reduction(primes, max_in_degree = float('inf')):
    """Implements the reduction method of Veliz-Cuba (2011).
    Deletion order is such that nodes with low in-degree are prioritized for
    removal. Deletion proceeds until all remaining nodes have self-loops.

    Parameters
    ----------
    primes : pyboolnet primes dictionary
        Update rules.
    max_in_degree : int or float
        Will not try to delete nodes with in-degree larger than this. Deleting
        nodes with large in-degree can be computationally expensive (the default
        is float('inf')).

    Returns
    -------
    reduced : pyboolnet primes dictionary
        The reduced primes.
    constants : partial state dictionary
        Node states that became logically fixed during reduction.

    """

    reduced, constants = remove_outdag(primes)
    G = pyboolnet.interaction_graphs.primes2igraph(reduced)
    cur_order = sorted(reduced,key=lambda x: G.in_degree(x))

    change = True
    while change and len(reduced) > 0:
        change = False
        for node in cur_order:
            retry_node = True
            if not node in reduced or G.in_degree(node) > max_in_degree:
                continue
            elif not any(node in p for p in reduced[node][1]):
                reduced, nc = delete_node(reduced, node)
                constants.update(nc)
                if len(reduced) > 0:
                    G = pyboolnet.interaction_graphs.primes2igraph(reduced)
                else:
                    G = nx.DiGraph()
                change = True
                # print("\n\n\nNEW PRIMES")
                # sm_format.pretty_print_prime_rules(reduced)
                break
        cur_order = sorted(reduced,key=lambda x: G.in_degree(x))

        pyboolnet.prime_implicants.percolate(reduced, add_constants=constants, remove_constants=False, copy=False)
        constants.update(pyboolnet.prime_implicants.find_constants(reduced))
        pyboolnet.prime_implicants.remove_all_constants(reduced, copy=False)
    return reduced, constants

def mediator_reduction(primes):
    """Network reduction method of Saadadtpour, Albert, Reluga (2013)
    Preserves fixed points. Number of complex attractors is often, but
    not always conserved (despite inital claims). Can be viewed as a more
    restrictive version of the deletion reduction method of Veliz-Cuba (2011).

    Parameters
    ----------
    primes : pyboolnet primes dictionary
        Update rules.

    Returns
    -------
    reduced : pyboolnet primes dictionary
        The reduced primes.
    constants : partial state dictionary
        Node states that became logically fixed during reduction.

    """
    return primes, {}
    reduced, constants = remove_outdag(primes)
    cur_order = sorted(reduced)
    G = pyboolnet.interaction_graphs.primes2igraph(reduced)
    candidates = [v for v in reduced if G.in_degree(v) == G.out_degree(v) == 1 and not G.has_edge(v,v)]

    for node in candidates:
        u = list(G.predecessors(node))[0]
        w = list(G.successors(node))[0]
        if not w in G.successors(u) and not w in G.predecessors(u):
            reduced, nc = delete_node(reduced, node)
            constants.update(nc)
            G = pyboolnet.interaction_graphs.primes2igraph(reduced)
            candidates = [v for v in reduced if G.in_degree(v) == G.out_degree(v) == 1 and not G.has_edge(v,v)]

    return reduced, constants


class MotifReduction:
    """Class to generate and store data about a network reduction that arises
    during the stable motif succession diagram construction algorithm.

    Parameters
    ----------
    motif_history : list of partial state dictionaries
        Stable motifs that can lock in to give the reduced network (in order).
    fixed : partial state dictionary
        Nodes values that have been fixed and reduced by stable motifs and their
        logical domain of influence.
    reduced_primes : pyboolnet primes dictionary
        Update rules for the reduced network.
    max_simulate_size : int
        Maximum number of variables for which to brute-force build a state
        transition graph (the default is 20).
    prioritize_source_motifs : bool
        Whether source nodes should be considered first (the default is True).
    max_stable_motifs : int
        Maximum number of output lines for pyboolnet to process from the
        AspSolver (the default is 10000).
    MPBN_update : bool
        Whether MPBN update is used instead of general asynchronous update
        (see Pauleve et al. 2020)(the default is False).

    Attributes
    ----------
    merged_history_permutations : list of lists of int
        Permutations of motif_history (by index) that also yield this reduction.
    logically_fixed_nodes : partial state dictionary
        Nodes values that have been fixed and reduced by stable motifs and their
        logical domain of influence.
    time_reverse_primes : pyboolnet primes dictionary
        Update rules of the time reversed reduced system.
    stable_motifs : list of partial state dictionaries
        Stable motifs of the reduced system.
    time_reverse_stable_motifs : list of partial state dictionaries
        Stable motifs of the time reversed system.
    merged_source_motifs : list of partial state dictionaries
        List of source-like stable motifs that have been merged into a single
        motif to avoid redundancy.
    source_independent_motifs : list of partial state dictionaries
        Stable motifs that exist independent of the values of the source nodes
    merge_source_motifs : list of partial state dictionaries
        Stable motifs generated by merging the stable motifs corresponding to
        source nodes.
    rspace : rspace list
        The rspace, or "restrict space" of the reduced network, describing a
        necessary condition for the system to avoid activating additional
        stable motifs (see restrict_space.py for further details).
    motif_history : list of partial state dictionaries
        Stable motifs that can lock in to give the reduced network (in order)
    reduced_primes : pyboolnet primes dictionary
        Update rules for the reduced network.
    fixed_rspace_nodes : partial state dictionary
        Nodes values that are fixed in the rspace.
    rspace_constraint : str
        BNET expression that is true in and only in the rspace.
    reduced_rspace_constraint : str
        S simplification of the rspace_constraint given the fixed_rspace_nodes
        states are satisfied
    rspace_update_primes : pyboolnet primes dictionary
        The update rules obtained from simplifying under the assumption that the
        fixed_rspace_nodes are fixed
    conserved_functions : list of pyboolnet expressions
        Boolean functions that are constant within every attractor, in pyboolnet
        update rule format
    rspace_attractor_candidates : list of str
        Attractors (lists of statestrings) in the rspace_update_primes that
        satisfy the reduced_rspace_constraint
    partial_STG : networkx.DiGraph
        Subgraph of the state transition graph of the reduced network that
        contains any and all attractors that do not lie in any of the reduced
        network's stable motifs.
    no_motif_attractors : list of str
        Complex attractors that do not "lock in" any additional stable motifs,
        stored as collections of state strings.
    attractor_dict_list : list of dictionaries
        Dictionaries corresponding to attractors that are in this reductions, but
        not in any of its subreductions (if it has any). Each describes the node
        states in the attractors according to the following
        1 variable is "ON"
        0 variable is "OFF"
        X variable is known to oscillate
        ? at least one such variable must oscillate
        ! the attractor may be false; if it is genuine, at least
          one such variable must oscillate
    terminal : str
        One of "yes", "no", or "possible", indicating whether the reduction contains
        attractors that are not in any of its subreductions.
    delprimes : pyboolnet prime dictionary
        Update rules for the system's deletion projection. Steady states and stable
        motif activation are preserved. These rules may yield additional, spurious
        complex attractors.
    deletion_STG : networkx.DiGraph
        Portion of the deletion projection's STG that contains all motif-avoidant
        attractors.
    deletion_no_motif_attractors : list of str
        Motif avoidant attractors of the deletion projection. The number of these is
        an upper bound on the number of motif avoidant attractors in the reduction.
    """

    def __init__(self,motif_history,fixed,reduced_primes,max_simulate_size=20,prioritize_source_motifs=True,max_stable_motifs=10000,MPBN_update=False):
        if motif_history is None:
            self.motif_history = []
        else:
            self.motif_history = motif_history.copy()
        self.merged_history_permutations = []
        self.logically_fixed_nodes = fixed
        self.reduced_primes = reduced_primes.copy()
        self.stable_motifs = pyboolnet.trap_spaces.compute_trap_spaces(self.reduced_primes, "max",max_output=max_stable_motifs)
        if MPBN_update==False:
            self.time_reverse_primes =sm_time.time_reverse_primes(self.reduced_primes)
            self.time_reverse_stable_motifs = pyboolnet.trap_spaces.compute_trap_spaces(self.time_reverse_primes, "max",max_output=max_stable_motifs)
        self.merged_source_motifs=None
        self.source_independent_motifs=None
        if self.motif_history == [] and prioritize_source_motifs:
            if MPBN_update==False:
                self.merge_source_motifs()
            else:
                self.simple_merge_source_motifs(reduced_primes,MPBN_update=MPBN_update)

        # These may or may not get calculated.
        # Sensible default values are in comments, but we will just use None for now.
        self.fixed_rspace_nodes=None # {}
        self.rspace_constraint=None # ""
        self.reduced_rspace_constraint=None # ""
        self.rspace_update_primes=None # {}
        self.conserved_functions=None # [[{}]]
        self.rspace_attractor_candidates=None # []
        self.partial_STG=None # nx.DiGraph()
        self.no_motif_attractors=None # []
        self.deletion_STG = None
        self.deletion_no_motif_attractors = None
        self.attractor_constants = None
        self.attractor_dict_list=None # []
        # self.attractor_dict={}

        # skips finding motif avoidant attractors when using MPBN_update
        if MPBN_update:
            if len(self.stable_motifs) == 0: # terminal reduction iff no more stable motifs
                self.terminal = "yes"
            else:
                self.terminal = "no"
            self.attractor_dict_list = self.generate_attr_dict_list(MPBN_update=MPBN_update)
            return

        self.rspace = sm_rspace.rspace(self.stable_motifs, self.time_reverse_stable_motifs,self.reduced_primes)

        study_possible_oscillation = False

        if not self.merged_source_motifs is None:
            self.terminal = "no"
        elif self.rspace == [[{'0':1}]] and len(self.stable_motifs) > 0: # a stable motif must lock in
            self.terminal = "no"
        elif self.rspace == [[{}]] and len(self.stable_motifs) > 0: # could not find 1-node drivers
            self.terminal = "possible"
            study_possible_oscillation = True
        elif len(self.stable_motifs) == 0: # necessarily terminal
            self.terminal = "yes"
            if len(self.reduced_primes) > 0: # Terminates in oscillation, else, fixed point
                study_possible_oscillation = True
        else: # found 1-node drivers, so we can investigate further
            self.terminal = "possible" # TODO: implement case-checking based on rspace

            #self.fixed_rspace_nodes =sm_rspace.fixed_rspace_nodes(self.rspace,self.reduced_primes)
            self.rspace = sm_rspace.reduce_rspace(self.rspace,self.reduced_primes)
            self.fixed_rspace_nodes = self.rspace[0][0]

            for motif in self.stable_motifs:
                if motif.items() <= self.fixed_rspace_nodes.items():
                    self.terminal = "no"
                    break
            if self.terminal == "possible":
                self.rspace_constraint = sm_format.pretty_print_rspace(self.rspace,simplify=False)
                #self.reduced_rspace_constraint = sm_rspace.reduce_rspace_string(self.rspace_constraint,self.fixed_rspace_nodes,simplify=False)
                self.reduced_rspace_constraint=sm_format.pretty_print_rspace(self.rspace[1:],simplify=False)
                self.rspace_update_primes = reduce_primes(self.fixed_rspace_nodes,self.reduced_primes)[0]
                #self.test_rspace(search_partial_STGs = search_partial_STGs)
            study_possible_oscillation = self.terminal == "possible" # value may be changed by test_rspace

        if study_possible_oscillation and max_simulate_size > 0:
            # Now, we check to see if we can afford to simulate the proper STG
            # to actually find the attractors. If we can't, we'll simulate the
            # reduction, which will give bounds on the number of motif-free
            # complex attractors.
            if self.rspace_update_primes is not None:
                simulate_size = len(self.rspace_update_primes)
            else:
                simulate_size = len(self.reduced_primes)

            if simulate_size < max_simulate_size:
                self.find_no_motif_attractors()
                if len(self.no_motif_attractors) == 0:
                    self.terminal = "no"
                else:
                    self.terminal = "yes"
                    self.conserved_functions = sm_rspace.attractor_space_candidates(self.stable_motifs,
                                                                         self.time_reverse_stable_motifs)
            else:
                #print("STG is too large to simulate ("+
                #str(simulate_size)+"/"+str(max_simulate_size)+
                #"). We will attempt reduction methods. Increase max_simulate_size to force simulation.")
                self.delprimes, self.attractor_constants = mediator_reduction(self.reduced_primes)
                self.delprimes, nc = deletion_reduction(self.delprimes)
                self.attractor_constants.update(nc)

                # Before building any STGs, let's see if we've already identified
                # that a stable motif must stabilize based on the reduction.
                for sm in self.stable_motifs:
                    sat = True
                    for k,v in sm.items():
                        if (k,v) in self.attractor_constants.items():
                            continue
                        elif k not in self.delprimes:
                            continue
                        sat = False
                        break
                    if sat:
                        self.terminal = "no"
                        #print("The reduction indicates that the branch is not terminal. No need to simulate.")
                        self.attractor_dict_list = self.generate_attr_dict_list()
                        return
                if len(self.delprimes) < max_simulate_size:
                    #print("Simulating deletion reduction ("+str(len(self.delprimes))+" nodes)...")
                    self.find_deletion_no_motif_attractors(max_stable_motifs=max_stable_motifs)
                    if len(self.deletion_no_motif_attractors) == 0 and self.terminal != "yes":
                        self.terminal = "no"
                    # else:
                    #     self.terminal = "possible"
                else:
                    print("The STG is still too large ("+str(len(self.delprimes))+").")
                    print("Further analysis of this branch is needed.")
        self.attractor_dict_list = self.generate_attr_dict_list()

    def merge_source_motifs(self):
        """Merges stable motifs (and time-reversal stable motifs) that correspond to source nodes, e.g. A*=A, into combined motifs to
        avoid combinatorial explosion. For example, A*=A, B*=B, C*=C produces six motifs that can stabilize in 8 ways; without
        merging, these 8 combinations lead to 8*3!=48 successions because they can be considered in any order. This is silly because
        source nodes all stabilize simultaneously.

        We will assume that stable motifs and time reverse stable motifs have already been computed.

        Note that a source node in the forward time system is a source node in the time reverse system as well.
        This follows from A* = A => A- = ~(A*(A=~A)) = ~(~A) = A.

        If A* = A or X (i.e., A=1 is a stable motif), then A- = ~(~A | X) = A & ~X, so A=0 is a time-reverse stable motif. A similar
        argument applies for the A=0 stable motif. Thus, a motif is only a source motif if it is also a time-reverse motif.
        """
        source_motifs = [x for x in self.stable_motifs if len(x) == 1 and x in self.time_reverse_stable_motifs]
        if source_motifs == []:
            return
        self.source_independent_motifs = [x for x in self.stable_motifs if not x in source_motifs]

        source_vars = list(set([next(iter(x.keys())) for x in source_motifs])) # a list of source nodes

        self.merged_source_motifs = []
        for state in it.product([0,1],repeat=len(source_vars)):
            self.merged_source_motifs.append({v:x for v,x in zip(source_vars,state)})

    def simple_merge_source_motifs(self,primes,MPBN_update=False):
        """Merges stable motifs (and time-reversal stable motifs) that correspond to source nodes, e.g. A*=A, into combined motifs to
        avoid combinatorial explosion. For example, A*=A, B*=B, C*=C produces six motifs that can stabilize in 8 ways; without
        merging, these 8 combinations lead to 8*3!=48 successions because they can be considered in any order. This is silly because
        source nodes all stabilize simultaneously.

        Assumes that stable_motifs have already been computed,
        but time_reverse_primes and time_reverse_stable_motifs are not.

        To be used in the case of MPBN update.

        Parameters
        ----------
        primes : pyboolnet primes dictionary
            pyboolnet update rules whose source node stable motifs are to be merged.

        MPBN_update : bool
            Whether MPBN update is used instead of general asynchronous update
            (see Pauleve et al. 2020)(the default is False).

        Returns
        -------
        self.source_independent_motifs : list of dictionaries
            list of stable motifs that are not source motifs
            [{'node1':bool,'node2':bool, ...}, {'node3':bool,'node4':bool, ...}, ...]
        self.merged_source_motifs : list of dictionaries
            list of group of source motifs fixed at the same time
            [{'source_node1':bool,'source_node2':bool, ...}, ...]

        """
        assert MPBN_update == True, "This function is for MPBN update only"

        # source nodes will have update rule of the form 'A':[[{'A':0}],[{'A':1}]]
        source_vars = []
        for x in primes.keys():
            if primes[x] == [[{x:0}],[{x:1}]]:
                source_vars.append(x)

        source_motifs = []
        for x in source_vars:
            source_motifs.append({x:0})
            source_motifs.append({x:1})

        if source_motifs == []:
            return

        self.source_independent_motifs = [x for x in self.stable_motifs if not x in source_motifs]

        self.merged_source_motifs = []
        for state in it.product([0,1],repeat=len(source_vars)):
            self.merged_source_motifs.append({v:x for v,x in zip(source_vars,state)})

    # def test_rspace(self, search_partial_STGs=True):
    #     STG=pyboolnet.state_transition_graphs.primes2stg(self.rspace_update_primes,"asynchronous")
    #     steady_states,complex_attractors=pyboolnet.attractors.compute_attractors_tarjan(STG)
    #     names = sorted(self.rspace_update_primes)
    #     attractors = complex_attractors+[[s] for s in steady_states]
    #     self.build_rspace_attractor_candidates(attractors)

    # def build_rspace_attractor_candidates(self,attractors):
    #     self.rspace_attractor_candidates = []
    #     for attractor in attractors:
    #         possible_rspace_attractor = True
    #         for state in attractor:
    #             # state_dict = {** sm_format.statestring2dict(state,names),**self.fixed_rspace_nodes}
    #             # if pyboolnet.boolean_logic.are_mutually_exclusive(self.rspace_constraint,
    #             #                                                  sm_format.implicant2bnet(state_dict)):
    #             if sm_rspace.partial_state_contradicts_rspace(sm_format.statestring2dict(state,names),self.rspace):
    #                 possible_rspace_attractor = False
    #                 break
    #         if possible_rspace_attractor:
    #             self.rspace_attractor_candidates.append(attractor)
    #
    #     if len(self.rspace_attractor_candidates) == 0:
    #         self.terminal = "no"

    def build_K0(self):
        """Helper function for smart STG building. Builds initial set of nodes
        that are not part of any motif-avoidant attractor.

        Returns
        -------
        set of str
            Statestrings that do not belong to any motif-avoidant attractor.

        """
        K = set()
        for sm in self.stable_motifs:
            fill_vars = [k for k in self.reduced_primes if not k in sm]
            for fills in it.product(['0','1'],repeat = len(fill_vars)):
                s = ''
                fi = 0
                for k in self.reduced_primes:
                    if k in sm:
                        s += str(sm[k])
                    else:
                        s += fills[fi]
                        fi += 1
                K.add(s)
        return K

    def in_motif(self,ss,names):
        """Tests whether the (partial) state ss is in any stable motifs

        Parameters
        ----------
        ss : str
            Statestring (possibly on a subspace).
        names : list of str
            Variable names ordered to correspond to the positions of ss.

        Returns
        -------
        bool
            Whether ss is in any stable motif of the reduced system.

        """
        for sm in self.stable_motifs:
            smin = True
            for i,r in enumerate(names):
                if r in sm and not int(ss[i]) == sm[r]:
                    smin = False
            if smin: return True
        return False

    def build_inspace(self,ss,names, tr_stable_motifs = None):
        """Helper function for smart STG building. List all time reversal stable
        motifs to which (partial) state ss belongs.

        Parameters
        ----------
        ss : str
            Statestring (possibly on a subspace).
        names : list of str
            Variable names ordered to correspond to the positions of ss.
        tr_stable_motifs : list of partial state dictionaries
            Time reverse stable motifs. If None, use all time reverse stable
            motifs in the reduced system (the default is None).

        Returns
        -------
        list of partial state dictionaries
            Time reverse stable motifs that are active in the state ss.

        """
        inspaces = []
        if tr_stable_motifs is None:
            tr_stable_motifs = self.time_reverse_stable_motifs

        for ts in tr_stable_motifs:
            tsin = True
            for i,r in enumerate(names):
                if r in ts and not int(ss[i]) == ts[r]:
                    tsin = False
            if tsin: inspaces.append(ts)
        return inspaces

    def build_deletion_STG(self,max_stable_motifs=10000):
        """Build a piece of the STG that is guaranteed to contain all
        motif-avoidant attractors of the deletion projection. Complex attractors
        found here may be spurious.

        Parameters
        ----------
        max_stable_motifs : int
            Maximum number of output lines for pyboolnet to process from the
            AspSolver (the default is 10000).

        """
        names = sorted(self.delprimes)
        name_ind = {n:i for i,n in enumerate(names)}
        trprimes = sm_time.time_reverse_primes(self.delprimes)
        trsms = pyboolnet.trap_spaces.compute_trap_spaces(trprimes,"max",max_output=max_stable_motifs)

        if self.rspace_update_primes is not None:
            delrnames = [x for x in sorted(self.rspace_update_primes) if x in self.delprimes]
            rname_ind = {n:i for i,n in enumerate(names) if n in delrnames}
            fixed = {k:v for k,v in self.fixed_rspace_nodes.items()}
        else:
            rnames = names.copy()
            rname_ind = name_ind.copy()
            fixed = {}
        sim_names = [x for x in names if not x in fixed]

        #K = self.build_K0()
        K = set()
        self.deletion_STG = nx.DiGraph()

        inspace_dict = {}
        t = 0
        T = 1
        # note: product order gives s counting up in binary from 00..0 to 11..1
        for s in it.product(['0','1'],repeat=len(sim_names)):
            sl = ['']*len(names)
            j = 0
            for i in range(len(names)):
                if names[i] in fixed:
                    sl[i] = str(fixed[names[i]])
                else:
                    sl[i] = s[j]
                    j += 1

            ss = ''.join(sl)

            if ss in K: continue
            if self.in_motif(ss,names): continue

            simstate = True

            inspace = self.build_inspace(ss,names,tr_stable_motifs = trsms)
            inspace_dict[ss] = inspace

            self.deletion_STG.add_node(ss) # might end up removing the node later
            for i,r in enumerate(names):
                nri = int(not int(ss[i]))
                # if any p below is satisfied, we get a change of state
                # the value of the new r will be equal to nri
                for p in self.delprimes[r][nri]:
                    psat = True
                    for k,v in p.items():
                        if not int(ss[name_ind[k]]) == v:
                            psat = False
                            break
                    if psat: # state change verified
                        child_state_list = list(ss)
                        child_state_list[i] = str(nri)
                        child_state = ''.join(child_state_list)

                        # Check if changed something that should be fixed or landed in K
                        # If not, check if we left a TR stable motif
                        prune = r in fixed or child_state in K
                        # next we check to see if we've left the rspace
                        # note that we don't have to check rspace[0], as this
                        # is handled by checking r in fixed
                        if not prune:
                            stdict = sm_format.statestring2dict(child_state,names)
                            prune = not sm_rspace.partial_state_contradicts_rspace(stdict, self.rspace[1:])

                        # next, we check to see if we left a TR motif
                        if not prune:
                            if not child_state in inspace_dict:
                                inspace_dict[child_state] = self.build_inspace(child_state,names,tr_stable_motifs = trsms)
                            prune = not inspace_dict[child_state] == inspace


                        # By here, prune is TRUE if we left a TR motif or are in K
                        if prune:
                            # prune the STG and stop simulating ss
                            simstate = False
                            rnodes = list(nx.bfs_tree(self.deletion_STG,ss,reverse=True).nodes())
                            K.update(rnodes)
                            self.deletion_STG.remove_nodes_from(rnodes)
                        else:
                            self.deletion_STG.add_edge(ss,child_state)
                        break # we know the ss at r changed, no need to check more primes
                if not simstate: break # don't check other vars: already found ss -> K

    def find_deletion_no_motif_attractors(self,max_stable_motifs=10000):
        """Identify motif-avoidant attractors in the deletion projection.

        Parameters
        ----------
        max_stable_motifs : int
            Maximum number of output lines for pyboolnet to process from the
            AspSolver (the default is 10000).

        """
        if self.deletion_STG is None:
            self.build_deletion_STG(max_stable_motifs=max_stable_motifs)

        # Note: fixed points of the deletion system are fixed points of the
        # undeleted system, so we ignore these as they must contain stable motifs
        if len(list(self.deletion_STG.nodes())) > 0:
            candidates = [x for x in nx.attracting_components(self.deletion_STG) if len(x) > 1]
        else:
            candidates = []
        self.deletion_no_motif_attractors = []

        # next, we see if any of these activate stable motifs
        names = sorted(self.delprimes)
        for att in candidates:
            no_motif = True
            for s in att:
                # The following check stems from the result that a stable motif
                # is active in an attractor of the original system iff its projection
                # is active in the projected attractor in the deletion-reduced system
                st = sm_format.statestring2dict(s,names)
                st.update(self.attractor_constants)
                if any(not sm_doi.fixed_excludes_implicant(st,sm) for sm in self.stable_motifs):
                    no_motif = False
                    break

            if no_motif:
                self.deletion_no_motif_attractors.append(att)

    def build_partial_STG(self):
        """Build a piece of the STG that is guaranteed to contain all
        motif-avoidant attractors of the reduction.

        """
        names = sorted(self.reduced_primes)
        name_ind = {n:i for i,n in enumerate(names)}

        if self.rspace_update_primes is not None:
            rnames = sorted(self.rspace_update_primes)
            rname_ind = {n:i for i,n in enumerate(names) if n in rnames}
            fixed = self.fixed_rspace_nodes
        else:
            rnames = names.copy()
            rname_ind = name_ind.copy()
            fixed = {}

        sim_names = [x for x in names if not x in fixed]


        #K = self.build_K0()
        K = set()
        self.partial_STG = nx.DiGraph()

        inspace_dict = {}
        t = 0
        T = 1
        # note: product order gives s counting up in binary from 00..0 to 11..1
        for s in it.product(['0','1'],repeat=len(sim_names)):
            sl = ['']*len(names)
            j = 0
            for i in range(len(names)):
                if names[i] in fixed:
                    sl[i] = str(fixed[names[i]])
                else:
                    sl[i] = s[j]
                    j += 1

            ss = ''.join(sl)

            if ss in K: continue
            if self.in_motif(ss,names): continue

            simstate = True

            inspace = self.build_inspace(ss,names)
            inspace_dict[ss] = inspace

            self.partial_STG.add_node(ss) # might end up removing ss later
            for i,r in enumerate(names):
                nri = int(not int(ss[i]))
                # if any p below is satisfied, we get a change of state
                # the value of the new r will be equal to nri
                for p in self.reduced_primes[r][nri]:
                    psat = True
                    for k,v in p.items():
                        if not int(ss[name_ind[k]]) == v:
                            psat = False
                            break
                    if psat: # state change verified
                        child_state_list = list(ss)
                        child_state_list[i] = str(nri)
                        child_state = ''.join(child_state_list)

                        # Check if changed something that should be fixed or landed in K
                        # If not, check if we left a TR stable motif
                        prune = r in fixed or child_state in K

                        # next we check to see if we've left the rspace
                        # note that we don't have to check rspace[0], as this
                        # is handled by checking r in fixed
                        if not prune:
                            stdict = sm_format.statestring2dict(child_state,names)
                            prune = not sm_rspace.state_in_rspace(stdict, self.rspace[1:])

                        # next, we check to see if we left a TR motif
                        if not prune:
                            if not child_state in inspace_dict:
                                inspace_dict[child_state] = self.build_inspace(child_state,names)
                            prune = not inspace_dict[child_state] == inspace


                        # By here, prune is TRUE if we left a TR motif or are in K
                        if prune:
                            # prune the STG and stop simulating ss
                            simstate = False
                            rnodes = list(nx.bfs_tree(self.partial_STG,ss,reverse=True).nodes())
                            K.update(rnodes)
                            self.partial_STG.remove_nodes_from(rnodes)
                        else:
                            self.partial_STG.add_edge(ss,child_state)
                        break # we know the ss at r changed, no need to check more primes
                if not simstate: break # don't check other vars: already found ss -> K

    def find_no_motif_attractors(self):
        """Find attractors of the reduction that are not present in any of its
        subreductions.

        """
        if self.partial_STG is None:
            self.build_partial_STG()
        if len(list(self.partial_STG.nodes())) > 0:
            self.no_motif_attractors = list(nx.attracting_components(self.partial_STG))
        else:
            self.no_motif_attractors = []

    def generate_attr_dict_list(self, MPBN_update=False):
        """Generate a list of attractors that are present in the reduction, but
        not in any of its subreductions.

        Parameters
        ----------
        MPBN_update : bool
            Whether MPBN update is used instead of general asynchronous update
            (see Pauleve et al. 2020)(the default is False).

        Returns
        -------
        list of dictionaries
            Dictionaries corresponding to attractors that are in this reductions, but
            not in any of its subreductions (if it has any). Each describes the node
            states in the attractors according to the following
            1 variable is "ON"
            0 variable is "OFF"
            X variable is known to oscillate
            ? at least one such variable must oscillate
            ! the attractor may be false; if it is genuine, at least
              one such variable must oscillate
        """
        attractors_dict={}

        #the reduction is not terminal --> no attractor
        if self.terminal == 'no':
            return []#'not terminal reduction' #I should replace this with an empty dict

        nodes_sorted = sorted(list(set(self.logically_fixed_nodes.keys()) | set(self.reduced_primes.keys()))) #steady state

        node_state_dict = self.logically_fixed_nodes.copy()
        if self.fixed_rspace_nodes is not None:
            node_state_dict.update(self.fixed_rspace_nodes)

        # Found a steady state (will always be terminal)
        if len(node_state_dict) == len(nodes_sorted):
            assert self.terminal == 'yes', "Found non-terminal steady state. This should not be possible!"
            node_state_dict = {k:v for k,v in sorted(node_state_dict.items())}
            return [node_state_dict]
        #non_fixed_nodes = sorted(list(set(nodes_sorted)-set(node_state_dict.keys())))
        non_fixed_nodes = [x for x in nodes_sorted if x not in node_state_dict]

        #the reduction is only possibly terminal
        if self.terminal=='possible':
            for n in non_fixed_nodes: #non-stabilized nodes
                node_state_dict[n] = '!'
                node_state_dict = {k:v for k,v in sorted(node_state_dict.items())}
            return [node_state_dict]

        #the reduction is definitely terminal
        elif self.terminal=='yes':
            #the reduction is terminal, not all nodes are fixed
            #and it is MPBN update.
            if MPBN_update == True:
                for n in non_fixed_nodes:
                    node_state_dict[n]='X'
                node_state_dict = {k:v for k,v in sorted(node_state_dict.items())}
                return [node_state_dict]

            #the reduction is terminal, not all nodes are fixed
            #and there is NO complex attractor mapped out
            elif self.no_motif_attractors is None:
                for n in non_fixed_nodes:
                    node_state_dict[n]='?'
                node_state_dict = {k:v for k,v in sorted(node_state_dict.items())}
                return [node_state_dict]

            #the reduction is terminal, not all nodes are fixed
            #there are complex attractors mapped out:
            attr_list=[]
            for complex_attractor in self.no_motif_attractors:
                ca_dict = sm_format.statelist2dict(non_fixed_nodes,complex_attractor)
                ns = node_state_dict.copy()
                ns.update(ca_dict)
                ns = {k:v for k,v in sorted(ns.items())}
                attr_list.append(ns)
                # # we check if there are stabilized nodes within the complex attractor
                # ca=self.find_constants_in_complex_attractor(complex_attractor)
                # node_state_dict=self.logically_fixed_nodes.copy()
                # for i in range(len(non_fixed_nodes)): #non-stabilized nodes
                #     node_state_dict[non_fixed_nodes[i]]=ca[i]
                #
                # attr_list.append(node_state_dict)
            return attr_list

    def find_constants_in_complex_attractor(self,c):
        """Given a set of strings representing the states of a complex attractor the function finds the nodes
        that are constant in the full complex attractor.

        Parameters
        ----------
        c : a set of binary strings
            Set of statestrings, e.g. set(['000', '010', '100']).

        Returns
        -------
        list of str
            An array consisting of 0s, 1s, and Xs. X represents an oscillating
            node, and the 0s and 1s represent nodes stabilized to those states.

        """
        import numpy as np
        ca=np.array([np.fromiter(i, int, count=len(i)) for i in c])
        attr=np.array(['X' for i in range(len(ca[0]))])
        sum_a0=ca.sum(axis=0)
        attr[np.where(sum_a0==0)[0]]=0
        attr[np.where(sum_a0==len(ca))[0]]=1
        return attr

    def summary(self,show_original_rules=True,hide_rules=False,show_explicit_permutations=False):
        """Print a summary of the reduction.

        Parameters
        ----------
        show_original_rules : bool
            Show rules of the unreduced system (the default is True)?
        hide_rules : bool
            Hide rules of the reduced system (the default is False)?
        show_explicit_permutations : bool
            Show motif permutations explicitly, instead of by index (the default is False)?

        """
        print("Motif History:",self.motif_history)
        print()
        print("Logically Fixed Nodes:",self.logically_fixed_nodes)
        print()
        if hide_rules:
            pass
        elif not self.motif_history == []:
            print("Reduced Update Rules:")
            sm_format.pretty_print_prime_rules(self.reduced_primes)
        else:
            if show_original_rules:
                print("Original Update Rules:")
                sm_format.pretty_print_prime_rules(self.reduced_primes)
            else:
                print("The update rules are not reduced.")
        print()
        if self.terminal == "no":
            if self.merged_source_motifs is None:
                print("At least one additional stable motif must stabilize.")
                print()
                print("Stable motifs:", self.stable_motifs)
            else:
                print("Source node values are not yet specified for the following nodes:",
                       ', '.join(sorted([k for k in self.merged_source_motifs[0]])))
                print()
                if self.source_independent_motifs == []:
                    print("There are no source-independent stable motifs.")
                else:
                    print("The following stable motifs exist independently of the source configuration:")
                    print(self.source_independent_motifs)
        elif self.terminal == "yes":
            if len(self.reduced_primes) > 0:
                print("There is a complex attractor in this reduced system in which no additional stable motifs activate.")
                print("At least some of the following must oscillate in such an attractor:")
                print(list(self.reduced_primes.keys()))
            else:
                print("This branch terminates in a steady state.")
        elif self.terminal == "possible":
            print("Some or none of these stable motifs may stabilize:",
                  self.stable_motifs)
            print()
            if not self.fixed_rspace_nodes is None:
                print("If no more stable motifs stabilize, these node states must be fixed:",
                      self.fixed_rspace_nodes)
                print()
                print("In addition, the following must stabilize to TRUE:")
                print(self.reduced_rspace_constraint)
                print()
                print("In this case, the unfixed nodes update according to the following rules:")
                sm_format.pretty_print_prime_rules(self.rspace_update_primes)

        if not self.conserved_functions is None:
            print()
            if len(self.conserved_functions) > 0:
                print("Found the following functions that are constant on attractors in this branch:")
                for x in self.conserved_functions:
                    if len(x) > 0:
                        sm_format.pretty_print_rspace([x],silent=False)
                        print()
            else:
                print("Unable to find non-trivial conserved functions for attractors in this branch.")
                print()
            if not self.no_motif_attractors is None:
                if len(self.no_motif_attractors) > 0:
                    print("Found the following complex attractors that do not lock in additional stable motifs:")
                    for x in self.no_motif_attractors:
                        print(x)

        if len(self.merged_history_permutations) > 0:
            print()
            print("This branch contains the following motif_history permutation(s):")
            if show_explicit_permutations:
                for x in self.merged_history_permutations: print([self.motif_history[i] for i in x])
            else:
                for x in self.merged_history_permutations: print(x)



import pystablemotifs as sm
import networkx as nx

def _format_reduction_label(s):
    """Helper function to make graph labels more readable. Removes single quotes
    and square brackets from the input string.

    Parameters
    ----------
    s : str
        Input label.

    Returns
    -------
    str
        Label with characters ', [, and ] removed.

    """
    return s.replace("'","").replace('[','').replace(']','')

def expanded_network(primes, single_parent_composites = False):
    """Produce the expanded network for given input update rules.

    Parameters
    ----------
    primes : pyboolnet primes dictionary
        The update rules for which to construct the expanded network.
    single_parent_composites : bool
        Whether to insert composite nodes between virtual nodes when one is a prime
        implicant of the other. If False, the number of nodes is decreased; if
        True, then the expanded network is bipartite (the default is False).

    Returns
    -------
    networkx.DiGraph
        Digraph representing the expanded network. Nodes have a 'type' attribute
        that can be either 'virtual' or 'composite'.

    """
    G = nx.DiGraph()
    cnode_id = 0
    for p in primes:
        for v in [0,1]:
            name = '('+str(p)+','+str(v)+')'
            G.add_node(name)
            G.nodes[name]['label'] = name
            G.nodes[name]['type'] = 'virtual'

            for hedge in primes[p][v]:
                G.add_node(cnode_id)
                G.nodes[cnode_id]['type'] = 'composite'
                G.add_edge(cnode_id,name)

                for k in hedge:
                    parent = '(' + str(k) + ',' + str(hedge[k]) + ')'
                    G.add_edge(parent,cnode_id)
                cnode_id += 1

    # If we want to remove composite nodes of "size" one
    if not single_parent_composites:
        for i in range(cnode_id):
            if G.in_degree(i) == 1:
                pre = list(G.predecessors(i))[0]
                suc = G.successors(i)
                for j in suc:
                    G.add_edge(pre,j)
                G.remove_node(i)


    return G

def networkx_succession_diagram(ar,include_attractors_in_diagram=True,use_compressed_diagram=True):
    """Label the succesion diagram and (optionally) attractors of the input attractor
    repertoire according to the conventions of Rozum et al. (2021). This is an
    alias for the function export.networkx_succession_diagram_reduced_network_based.

    Parameters
    ----------
    ar : AttractorRepertoire
        Attractor repertoire object for which to build the diagram.
    include_attractors_in_diagram : bool
        Whether attractors should be represented as nodes in the diagram (the
        default is True).
    use_compressed_diagram : bool
        Whether to use the (potentially compressed) succession diagram stored in
        ar.succession_digraph instead of the complete one ar.succession_diagram.digraph.
        These are equivelent unless ar.simplify_diagram is called. See
        AttractorRepertoire.py for additional details. The default is True.

    Returns
    -------
    networkx.DiGraph
        A labeled digraph that represents the succession diagram.

    """
    return networkx_succession_diagram_reduced_network_based(ar,
        include_attractors_in_diagram=include_attractors_in_diagram,
        use_compressed_diagram=use_compressed_diagram)

def networkx_succession_diagram_reduced_network_based(ar,include_attractors_in_diagram=True,use_compressed_diagram=True):
    """Label the succesion diagram and (optionally) attractors of the input attractor
    repertoire according to the conventions of Rozum et al. (2021).

    Parameters
    ----------
    ar : AttractorRepertoire
        Attractor repertoire object for which to build the diagram.
    include_attractors_in_diagram : bool
        Whether attractors should be represented as nodes in the diagram (the
        default is True).
    use_compressed_diagram : bool
        Whether to use the (potentially compressed) succession diagram stored in
        ar.succession_digraph instead of the complete one ar.succession_diagram.digraph.
        These are equivelent unless ar.simplify_diagram is called. See
        AttractorRepertoire.py for additional details. The default is True.

    Returns
    -------
    networkx.DiGraph
        A labeled digraph that represents the succession diagram.

    """

    if use_compressed_diagram:
        G_reduced_network_based=ar.succession_digraph.copy()
        relevant_nodes = ar.relevant_nodes
    else:
        G_reduced_network_based=ar.succession_diagram.digraph.copy()
        relevant_nodes = ar.primes.keys()

    has_nodes = False
    for i in G_reduced_network_based.nodes():
        has_nodes = True
        states={}
        for k,v in ar.succession_diagram.motif_reduction_dict[i].logically_fixed_nodes.items():
            if k in relevant_nodes:
                states[k]=v
        G_reduced_network_based.nodes[i]['index'] = i
        G_reduced_network_based.nodes[i]['states']=states

        contraction_indices = [i]
        if use_compressed_diagram and 'contraction' in ar.succession_digraph.nodes[i].keys():
            contraction_indices += list(ar.succession_digraph.nodes[i]['contraction'].keys())

        histories = []
        for c in contraction_indices:
            histories.append(ar.succession_diagram.motif_reduction_dict[c].motif_history)
        G_reduced_network_based.nodes[i]['history']=histories
        G_reduced_network_based.nodes[i]['motif union'] = {k:v for h in histories
            for m in h
            for k,v in m.items()
            if k in ar.relevant_nodes or not use_compressed_diagram}

    if not has_nodes:
        G_reduced_network_based.add_node(0)
        G_reduced_network_based.nodes[0]['index'] = 0
        G_reduced_network_based.nodes[0]['states'] = {}
        G_reduced_network_based.nodes[0]['history'] = [[]]
        G_reduced_network_based.nodes[0]['motif union'] = {}

    for u,v in G_reduced_network_based.edges():
        ufixed = G_reduced_network_based.nodes[u]['states']
        vfixed = G_reduced_network_based.nodes[v]['states']
        lock_in = {}
        for k,val in vfixed.items():
            if k not in ufixed.keys():
                lock_in[k] = val
        G_reduced_network_based.edges[u,v]['states'] = lock_in

        ufixed = G_reduced_network_based.nodes[u]['motif union']
        vfixed = G_reduced_network_based.nodes[v]['motif union']
        lock_in = {}
        for k,val in vfixed.items():
            if k not in ufixed.keys():
                lock_in[k] = val
        G_reduced_network_based.edges[u,v]['motif'] = lock_in

    if include_attractors_in_diagram and not use_compressed_diagram:
        for a_index,a in enumerate(ar.attractors):
            astr = 'A'+str(a_index)
            G_reduced_network_based.add_node(astr)
            G_reduced_network_based.nodes[astr]['index']=astr
            G_reduced_network_based.nodes[astr]['states']=a.attractor_dict

            for r in a.reductions:
                r_ind = list(ar.succession_diagram.motif_reduction_dict.values()).index(r)
                r_key=list(ar.succession_diagram.motif_reduction_dict.keys())[r_ind]
                if r_key in G_reduced_network_based.nodes():
                    G_reduced_network_based.add_edge(r_key,astr,states='')
    if include_attractors_in_diagram and use_compressed_diagram:
        for a_index,a in enumerate(ar.attractor_equivalence_classes):
            astr = 'A'+str(a_index)
            G_reduced_network_based.add_node(astr)
            G_reduced_network_based.nodes[astr]['index']=astr
            G_reduced_network_based.nodes[astr]['states']=a['states']

            for r_key in a['reductions']:
                if r_key in G_reduced_network_based.nodes():
                    G_reduced_network_based.add_edge(r_key,astr,states='')

    for n in G_reduced_network_based.nodes():
        if str(n)[0]=='A':
            G_reduced_network_based.nodes[n]['label']='Attractor ' +\
                G_reduced_network_based.nodes[n]['index'] +\
                ':\n '+str(G_reduced_network_based.nodes[n]['states'])
        else:
            G_reduced_network_based.nodes[n]['label']=str(G_reduced_network_based.nodes[n]['motif union'])

    return G_reduced_network_based

def networkx_succession_diagram_motif_based(ar,include_attractors_in_diagram=True):
    """Label the succesion diagram and (optionally) attractors of the input attractor
    repertoire according to the conventions of Zanudo and Albert (2015). If
    attractors are not included, this is the line graph of the succession diagram
    defined in Rozum et al. (2021). Does not support compression.

    Parameters
    ----------
    ar : AttractorRepertoire
        Attractor repertoire object for which to build the diagram.
    include_attractors_in_diagram : bool
        Whether attractors should be represented as nodes in the diagram (the
        default is True).

    Returns
    -------
    networkx.DiGraph
        A labeled digraph that represents the succession diagram.

    """
    G_reduced_network_based = networkx_succession_diagram_reduced_network_based(ar,
        include_attractors_in_diagram=False,
        use_compressed_diagram=False)

    G_motif_based = nx.line_graph(G_reduced_network_based)

    for e in G_reduced_network_based.edges():
        G_motif_based.nodes[e]['states'] = G_reduced_network_based.edges[e]['states']
        G_motif_based.nodes[e]['motif'] = G_reduced_network_based.edges[e]['motif']
        G_motif_based.nodes[e]['label'] = str(G_reduced_network_based.edges[e]['motif'])
    for e in G_motif_based.edges():
        G_motif_based.edges[e]['states']=''


    if include_attractors_in_diagram:
        for a_index,a in enumerate(ar.attractors):
            astr = 'A'+str(a_index)
            G_motif_based.add_node(astr)
            G_motif_based.nodes[astr]['states']=a.attractor_dict
            G_motif_based.nodes[astr]['label']='Attractor ' + astr +\
                ':\n '+ str(a.attractor_dict)
            for r in a.reductions:
                r_ind = list(ar.succession_diagram.motif_reduction_dict.values()).index(r)
                r_key=list(ar.succession_diagram.motif_reduction_dict.keys())[r_ind]
                for n in G_motif_based.nodes():
                    if type(n)==tuple:
                        i,j=n
                        if r_key==j:
                            G_motif_based.add_edge((i,j),astr,states='')
    return G_motif_based

def plot_nx_succession_diagram(G, pos=None, fig_dimensions=(None,None), nx_node_kwargs=None, nx_edge_kwargs=None,
    draw_node_labels=True, labeling_convention='label', draw_edge_labels=False, nx_node_label_kwargs=None, nx_edge_label_kwargs=None):
    """Plot the input succession diagram. Requires matplotlib. For finer control
    over plot appearance, it is recommended to plot g directly.

    Parameters
    ----------
    G : networkx.DiGraph
        Labeled succession diagram, e.g., as is output from
        export.networkx_succession_diagram_reduced_network_based().
    fig_dimensions : (int,int)
        Dimensions of the output figure. If (None,None), then the dimensions are
        calculated based on the number of nodes in g (the default is (None,None)).
    pos : str or graphviz_layout
        Layout for the nodes; A dictionary with nodes as keys and positions as
        values. Positions should be sequences of length 2. If none, we attempt to
        use pydot/graphviz to construct a layout, otherwise we fall back to the
        networkx planar_layout function (succession diagrams are always planar).
    draw_node_labels : bool
        Whether node labels should be drawn (True) or left as metadata (False)
        (the default is True).
    draw_edge_labels : bool
        Whether edge labels should be drawn (True) or left as metadata (False);
        only affects reduced-network-based (default) succession diagrams, not
        motif-based succession diagrams. (The default value is False.)
    labeling_convention : str
        Whether edge labels should be just the stable motifs ('label') or all stabilized states ('states')
        (the default is 'label').
    nx_node_kwargs : dictionary
        Keword arguments passed to nx.draw_networkx_nodes (in addition to G and pos).
        If None, we pass {'node_size':50*G.number_of_nodes()} by default.
    nx_edge_kwargs : dictionary
        Keword arguments passed to nx.draw_networkx_edges (in addition to G and pos).
        If None, we pass {'arrowstyle':'-|>','width':2,'arrowsize':30} by default.
    nx_node_label_kwargs : dictionary
        Keword arguments passed to nx.draw_networkx_labels (in addition to G and pos).
        If None, we pass {'font_size':16} by default.
    nx_edge_label_kwargs : dictionary
        Keword arguments passed to nx.draw_networkx_edge_labels (in addition to G and pos).
        If None, we pass {'font_size':16} by default.

    """
    import matplotlib.pyplot as plt

    if fig_dimensions == (None,None):
        fig_dimensions=(2*(G.number_of_nodes()+2),G.number_of_nodes()+2)

    if pos is None:
        try:
            from networkx.drawing.nx_agraph import graphviz_layout
            pos = graphviz_layout(G, prog='dot')
        except ImportError:
            pos = nx.planar_layout(G)

    plt.figure(figsize=fig_dimensions)

    if nx_node_kwargs is None:
        nx_node_kwargs = {'node_size':50*G.number_of_nodes()}
    if nx_edge_kwargs is None:
        nx_edge_kwargs = {'arrowstyle':'-|>','width':2,'arrowsize':30}
    if nx_node_label_kwargs is None:
        nx_node_label_kwargs = {'font_size':16}
    if nx_edge_label_kwargs is None:
        nx_edge_label_kwargs = {'font_size':16}

    nx.drawing.draw_networkx_nodes(G, pos,**nx_node_kwargs)
    nx.draw_networkx_edges(G, pos,**nx_edge_kwargs)
    if draw_node_labels:
        if labeling_convention=='label':
            nx.drawing.draw_networkx_labels(G,pos, labels=dict(G.nodes('label')),**nx_node_label_kwargs)
        else:
            nx.drawing.draw_networkx_labels(G,pos, labels=dict(G.nodes('states')),**nx_node_label_kwargs)
    if draw_edge_labels:
        nx.drawing.draw_networkx_edge_labels(G,pos,edge_labels=nx.get_edge_attributes(G,'motif'),**nx_edge_label_kwargs)
    plt.axis('off')
    plt.show()

def attractor_dataframe(ar):
    """Summarize the input attractor repertoire in a pandas DataFrame (requires
    pandas).

    Parameters
    ----------
    ar : AttractorRepertoire
        Attractor repertoire to summarize.

    Returns
    -------
    pandas.DataFrame
        Summary of the attractors.

    """
    import pandas as pd
    df=pd.DataFrame()
    for a in ar.attractors:
        df=df.append(a.attractor_dict,ignore_index=True).astype(int, errors='ignore').astype(str)

    return df

def save_to_graphml(G,model_name):
    """Export a labeled succesion diagram to graphml format.

    Parameters
    ----------
    G : networkx.DiGraph
        Labeled succession diagram to export.
    model_name : str
        Name of file to save to (.graphml extension will be appended).

    """

    #Graphml does not support complex attribues so we create a copy and stringify everything
    G_ex=G.copy()
    for n in G_ex.nodes():
        for k,v in G_ex.nodes[n].items():
            G_ex.nodes[n][k]=str(v)
    for e in G_ex.edges():
        for k,v in G_ex.edges[e].items():
            G_ex.edges[e][k]=str(v)
    nx.write_graphml(G_ex, "%s.graphml"%model_name)
